# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# yafe: Yet Another Framework for Experiments.
#
# Licence
# -------
# This file is part of yafe.
#
# yafe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Base functions and classes for experiments.

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
.. moduleauthor:: Florent Jaillet

Designing an :class:`~yafe.base.Experiment`
===========================================

Designing an experiment consists in

* Creating an :class:`~yafe.base.Experiment` object which includes, as
  arguments of the constructor:

    * a ``get_data`` function to load data,
    * a ``get_problem`` class or function to generate problems from the data,
    * a ``get_solver`` class or function to solve problems,
    * a ``measure`` function to compute performance.

* Adding tasks to the :class:`~yafe.base.Experiment` object by specifying the
  data, problem and solver parameters.

The API for designing an experiment is described below and is illustrated in
the :ref:`yafe tutorial </_notebooks/yafe.ipynb>`.

Loading data: ``get_data`` and data parameters
----------------------------------------------
Experiments are often applied to several data items.
Items may be entries loaded from a dataset or data synthesized from some
control parameters.

The input of function ``get_data`` is one or several parameters that
characterize which data item must be returned.

The output of function ``get_data`` is a dictionary with arbitrary keys and
whom values are the returned data.

Note that in the current version, ``get_data`` should have at least one
parameter (which may take a single value).

Examples:

* For a dataset of ``n`` items, one may design a ``get_data(i_item)`` function
  that takes an integer ``i_item``, load item number ``i_item`` in a variable
  ``x`` and returns a dictionary ``{'data_item': x}`` ; one may then add
  tasks using ``data_params = {'i_item': np.arange(n)}``.
* For synthetic data, one may design a ``get_data(f0)`` function
  that takes an real-valued frequency ``f0``, synthesize a sinusoid with
  frequency ``f0`` in a variable ``x`` and returns a dictionary
  ``{'signal': x}`` ; one may then add tasks using
  ``data_params = {'f0': np.linspace(0, 1, 10)}``.

Generating problems: ``get_problem`` and problem parameters
-----------------------------------------------------------
Problems to be solved are generated from the data. Several instances of a
problem may be generated from one data item, by varying some problem
parameters, e.g., amount of noise. Thus, designing the problem generation
stage consists in specifying how to turn problems parameters and data into
problem data (and solution). This may be done equivalently by a class or a
function.

``get_problem`` may be a class, in which case:

* the inputs of the ``__init__`` method are the parameters of the problem,
* the parameters of the ``__call__`` method must match the keys of the
  dictionary obtained from ``get_data``,
* the output of the ``__call__`` method is a tuple
  ``(problem_data, solution_data)`` where ``problem_data`` is a dictionary
  containing the problem data for the solver and ``solution_data`` is a
  dictionary containing the solution of the problem for the performance
  measure.

Thus, an instance of that class is generated for each set of problem
parameters and is called on data to generate the problem data and its solution.

Alternatively, ``get_problem`` may be a function, in which case:

* its inputs are the parameters of the problem,
* its output must be a function that takes some data as inputs and returns the
  tuple ``(problem_data, solution_data)`` as described in the case above.

Note that in the current version, ``get_problem`` should have at least one
parameter (which may take a single value).

Solving problems: ``get_solver`` and solver parameters
------------------------------------------------------
A solver generates a solution from some problem data. Several instances of a
solver may be used by varying the solver's parameters, e.g., the order of the
model. As for problems, solvers can be implemented equivalently by a
class or a function.

``get_solver`` may be a class, in which case:

* the inputs of the ``__init__`` method are the parameters of the solver,
* the parameters of the ``__call__`` method must match the keys of the
  dictionary obtained from the problem generation method ``__call__``,
* the output of the ``__call__`` method is a dictionary containing the solution
  generated by the solver.

Alternatively, ``get_solver`` may be a function, in which case:

* its inputs are the parameters of the solver,
* its output must be a function that takes some problem data as inputs and
  returns a dictionary containing the solution generated by the solver,
  similarly as the ``__call__`` method described above.

Note that in the current version, ``get_solver`` should have at least one
parameter (which may take a single value).

Computing performance: ``measure``
----------------------------------
Several performance measures may be calculated from the estimated solution,
using other data like the original data, the problem data, or parameters of
the data, problem or solver.

These performance measures must be computed within a *function* ``measure``
as follows:

* its arguments should be dictionaries ``source_data``, ``problem_data``,
  ``solution_data`` and ``solved_data``, as returned by the data access
  function, the ``__call__`` methods of problem generation class and the
  ``__call__`` methods of solver class, respectively; an additional argument
  ``task_params`` is a dictionary that contains the data, problem and solver
  parameters;
* its output is a dictionary whose keys and values are the names and values of
  the various performance measures.

Setting data, problem and solver parameters: adding tasks
---------------------------------------------------------

Once an instance of :class:`~yafe.base.Experiment` is created based on
the four processing blocks given by functions or classes  ``get_data``,
``get_problem``, ``get_solver`` and ``measure``, one must define the
parameters that must be used for each of these blocks. This is done by
calling method :meth:`~yafe.base.Experiment.add_tasks`: its inputs
are three dictionaries ``data_params``, ``problem_params`` and
``solver_params`` whose keys match the input of ``get_data``, ``get_problem``
and ``get_solver`` respectively, and whose values are array-like objects
containing the parameters that must be used in the instances of those
processing blocks. A hypercube is automatically generated, based on the
cartesian product of all the arrays of parameters (internally stored in a
so-called *schema*). Each point in this hypercube is a full set of
parameters to define a so-called *task* by generating some particular instances
of data, problem and solver.

Running an :class:`~yafe.base.Experiment`
=========================================

Once an :class:`~yafe.base.Experiment` is defined and tasks have been added,
one can run the tasks by calling method
:meth:`~yafe.base.Experiment.launch_experiment`, which executes all the
tasks, except those that may have already been executed previously. At any
time, one may control the execution of the tasks by calling method
:meth:`~yafe.base.Experiment.display_status`.

Running an :class:`~yafe.base.Experiment` on a cluster
------------------------------------------------------
Tasks may also be executed in parallel, individually or in batches, by
specifying the parameter ``task_ids`` in
:meth:`~yafe.base.Experiment.launch_experiment`. Function
:func:`~yafe.utils.generate_oar_script` may be used or taken as an example to
split the set of tasks among jobs that can be run by several cores in parallel.

Adding tasks
------------
Method :meth:`~yafe.base.Experiment.add_tasks` may be used to add new
parameter values, in which case the parameter hypercube is extended,
recomputing all cartesian products between parameters, and new tasks are
created while results from already-executed tasks are kept.

Collecting results
------------------
Method :meth:`~yafe.base.Experiment.collect_results` gathers the task results
into a single structure and saves it into a single file. Method
:meth:`~yafe.base.Experiment.load_results` then loads and returns the
results so that the user can process and display them. The main structure
that contains the results is a multi-dimensional array which dimensions match
the parameters of the tasks (see above) and the performance measures (last
dimension). This array may be loaded as an :class:`numpy.ndarray` or as a
:class:`xarray.DataArray`.

Misc
====

Tasks IDs and parameters
------------------------
Task IDs are generated from the parameters and depend on the sequence of
calls to method :meth:`~yafe.base.Experiment.add_tasks`. Finding a task ID
from a set of parameters and vice-versa is not trivial and may be done by
calling methods :meth:`~yafe.base.Experiment.get_task_data_by_id` and
:meth:`~yafe.base.Experiment.get_task_data_by_params`.

Dealing with several solvers
----------------------------
In order to compare several solvers, one may design one experiment by
solver, using the same generators and parameters for data and problems,
and merge the results.

Random seeds (reproducibility, consistent problem generation)
-------------------------------------------------------------
It is often important to control the random seeds, e.g., in order to reproduce
the experiments. Since all the tasks are run independently, controlling the
random seed is also required in order to have the exact same problem data and
solution in all tasks with the same data parameters and problem parameters,
especially when noise is added.

This can be obtained by using the *random state* mechanism provided by
:class:`numpy.random.RandomState`. The random state seed may be used inside
each block (see ``get_problem`` in
:ref:`yafe tutorial </_notebooks/yafe.ipynb>`) or passed as a parameter to
each processing block (as in ``scikit-learn``).

Files and folders
-----------------

A number of files and subfloders are generated in a folder named as the
experiment's name and located in the ``data_path`` set when creating the
experiment. The user may not need to handle those files since results are then
collected automatically, unless in case of tracking errors.

At the experiment level, a so-called *schema* is generated and contains all
the data, problem and solver parameters (file ``_schema.pickle``). Results
from all tasks are gathered in a single file ``results.npz``.

At the task level, a sub-directory is created for each task. It contains
several files generated when the task is created (parameters of the task in
``task_params.pickle``), when the task is run (raw data in
``source_data.pickle``, problem data in ``problem_data.pickle``,
true solution in ``solution_data.pickle``, solved data in
``solved_data.pickle``, performance measures in ``result.pickle``) and when
an error occurs during any processing step when running a task (``error_log``).

.. _yafe_configuration_file:

:mod:`yafe` configuration file
------------------------------

The :mod:`yafe` package uses a configuration file stored in
``$HOME/.config/yafe.conf`` to specify the default ``data_path`` used to store
the experiment data, as well as the default ``logger_path`` used to store
the processing log.

You can create an empty configuration file by running the static method
:meth:`~yafe.utils.ConfigParser.generate_config` and then edit the generated
file in ``$HOME/.config/yafe.conf``.

The resulting configuration file should have the following form::

    [USER]
    # user directory (data generated by experiments)
    data_path = /some/directory/path

    [LOGGER]
    # path to the log file
    path = /some/file/path/yafe.log

If you want to avoid the use of the :mod:`yafe` configuration file, you need
to systematically specify the ``data_path`` and ``logger_path`` parameters when
creating an :class:`~yafe.base.Experiment` (or alternatively for
``logger_path`` you can set the flag ``log_to_file`` to ``False``).
"""
from collections import OrderedDict
import inspect
import itertools
import logging
import os
from pathlib import Path
import pickle
import shutil
import sys
import traceback

import numpy as np

from .utils import get_logger, ConfigParser


class Experiment:
    """Definition of a class to design experiments.

    Parameters
    ----------
    name : str
        Name of the experiment, used to store experiment data in an eponym
        subfolder.
    get_data : function
        A function that returns data from some data parameters.
    get_problem : callable class or function
        A class or function that takes some problem parameters and returns a
        callable object that takes input data from method ``get_data`` and
        returns some problem data and true solution data.
    get_solver : callable object or function
        A class or function that takes some solver parameters and returns a
        callable object that takes problem data from method ``get_problem``
        and returns some solution data.
    measure : function
        Function that takes some task information and data returned by methods
        ``get_data``, ``get_problem``  and ``get_solver`` in order to compute
        one or several performance measures.
    force_reset : bool
        Flag forcing resetting the Experiment during initialization.
        If True, the previously computed data stored on disk for the
        Experiment with the same name are destroyed.
        If False, the previously computed data are reused (default option).
    data_path : None or str or pathlib.Path
        Path of the folder where the experiment data subfolder must be created.
        If ``data_path`` is ``None``, the data path given in the yafe
        configuration file is used.
    log_to_file : bool
        Indicates if the processing and debug information must be logged in a
        file.
    log_to_console : bool
        Indicates if the processing information must be displayed in the
        console.
    logger_path : None or str or pathlib.Path
        Full file path where the logged data must be written when
        ``log_to_file == True``.
        If ``logger_path`` is ``None`` the logger path given in the yafe
        configuration file is used.
    """
    def __init__(self, name, get_data, get_problem, get_solver, measure,
                 force_reset=False, data_path=None, log_to_file=True,
                 log_to_console=False, logger_path=None):

        # store attributes
        self.name = str(name)

        if data_path is None:
            data_path = ConfigParser().get_path('USER', 'data_path')
            self.xp_path = data_path / self.name
        else:
            self.xp_path = Path(data_path) / self.name

        self.get_data = get_data
        self.get_problem = get_problem
        self.get_solver = get_solver

        meas_params = inspect.signature(measure).parameters
        meas_expected_params = ('task_params', 'source_data', 'problem_data',
                                'solution_data', 'solved_data')
        for param in meas_expected_params:
            if param not in meas_params:
                msg = ('Wrong signature for measure. measure must accept the '
                       'following parameters: '
                       '{}.'.format(', '.join(meas_expected_params)))
                raise TypeError(msg)
        self.measure = measure

        self.logger = get_logger('xp_{}'.format(self.name), log_to_file,
                                 log_to_console, logger_path)

        if not self.xp_path.exists():
            self.logger.info('Init experiment %s', self.name)
            self._create_dirs_and_schema()
            return

        if force_reset:
            self.reset()
            return

        self.logger.info('Load experiment %s', self.name)
        if (self.xp_path / '_schema.pickle').exists():
            # Check that the existing schema is coherent with the signature
            # of the provided functions
            schema_keys = [
                ('data', par)
                for par in inspect.signature(self.get_data).parameters]
            schema_keys += [
                ('problem', par)
                for par in inspect.signature(self.get_problem).parameters]
            schema_keys += [
                ('solver', par)
                for par in inspect.signature(self.get_solver).parameters]
            if set(self._schema.keys()) != set(schema_keys):
                msg = ('The schema of the Experiment does not match the '
                       'schema of the Experiment with the same name stored in '
                       '{}.'. format(str(self.xp_path)))
                raise RuntimeError(msg)
        else:
            self.logger.info('No _schema.pickle file found, '
                             'resetting experiment %s', self.name)
            self.reset()

    def __str__(self):

        string = 'Name: {}\n'.format(self.name)
        string += 'Data stored in: {}\n'.format(self.xp_path)
        logger_path = None
        for handler in self.logger.handlers:
            if isinstance(handler, logging.FileHandler):
                logger_path = handler.baseFilename
                break
        if logger_path is None:
            string += 'Not storing any debug information\n'
        else:
            string += 'Debug information stored in: {}\n'.format(logger_path)
        string += 'Data provider: {}\n'.format(self.get_data)
        string += 'Problem provider : {}\n'.format(self.get_problem)
        string += 'Solver provider: {}\n'.format(self.get_solver)
        string += 'Measure: {}\n'.format(self.measure)
        string += '\n'
        string += 'Parameters:\n'
        data_param_str = ''
        problem_param_str = ''
        solver_param_str = ''
        for (cat, param_name), val in self._schema.items():
            if cat == 'data':
                data_param_str += '  {}: {}\n'.format(param_name, val)
            elif cat == 'problem':
                problem_param_str += '  {}: {}\n'.format(param_name, val)
            else:  # cat == 'solver':
                solver_param_str += '  {}: {}\n'.format(param_name, val)
        string += ' Data parameters:\n'
        string += data_param_str
        string += ' Problem parameters:\n'
        string += problem_param_str
        string += ' Solver parameters:\n'
        string += solver_param_str
        string += '\n'
        string += 'Status:\n'
        string += self._get_status(True)
        return string

    def reset(self):
        """Remove all tasks and previsouly computed results."""
        self.logger.info('Reset experiment %s', self.name)
        if self.xp_path.exists():
            shutil.rmtree(str(self.xp_path))
        self._create_dirs_and_schema()

    def _create_dirs_and_schema(self):
        self.xp_path.mkdir()
        (self.xp_path / 'tasks').mkdir()
        schema = {}
        for data_param in inspect.signature(self.get_data).parameters:
            schema[('data', data_param)] = []
        for problem_param in inspect.signature(self.get_problem).parameters:
            schema[('problem', problem_param)] = []
        for solver_param in inspect.signature(self.get_solver).parameters:
            schema[('solver', solver_param)] = []
        self._schema = schema

    @property
    def _schema(self):
        """Schema of the parameters for each section of the experiment.

        Parameters
        ----------
        _schema : dict
            Schema of parameters of the experiment.

        Notes
        -----
        The content is written in a pickle file. Use I/O with caution.
        """
        schema_path = self.xp_path / '_schema.pickle'
        if not schema_path.exists():
            raise FileNotFoundError(
                'Missing _schema file for xp {}'.format(self.name))
        else:
            with open(str(schema_path), 'rb') as fin:
                schema = pickle.load(file=fin)

        return schema

    @_schema.setter
    def _schema(self, schema):
        schema_path = self.xp_path / '_schema.pickle'
        with open(str(schema_path), 'wb') as fout:
            pickle.dump(schema, file=fout)

    # Task management (methods for internal use, should not be called by user)
    def _has_item(self, type_item, idt):
        """Indicates if an item of a given task exists or not.

        Parameters
        ----------
        type_item : str
            Type of requested item ({'task_params', 'source_data',
            'problem_data', 'solved_data', 'result'})
        idt : int
            Id of the item.

        Returns
        -------
        bool
            Return True if the item for the given task exists, otherwise
            False.
        """
        filename = type_item + '.pickle'
        path_item = self.xp_path / 'tasks' / '{:06}'.format(idt) / filename
        return path_item.exists()

    def _read_item(self, type_item, idt):
        """Load an item from the id.

        Parameters
        ----------
        type_item : str
            Type of requested item.
        idt : int
            Id of the item.

        Returns
        -------
        dict or None
        """
        path_task = self.xp_path / 'tasks' / '{:06}'.format(idt)
        if self._has_item(type_item, idt):
            filename = type_item + '.pickle'
            with open(str(path_task / filename), 'rb') as fin:
                return pickle.load(file=fin)
        else:
            return None

    def _write_item(self, type_item, idt, content):
        """Write an item with the id.

        Parameters
        ----------
        type_item : str
            Type of requested item.
        id : int
            Id of the item.
        content : str
            Content to write. Should be JSON serializable.
        """
        task_path = self.xp_path / 'tasks' / '{:06}'.format(idt)
        if not task_path.exists():
            task_path.mkdir()

        filename = type_item + '.pickle'
        with open(str(task_path / filename), 'wb') as fout:
            pickle.dump(content, file=fout)

    def _delete_item(self, idt, type_item=''):
        """Delete an item from the id of a task.

        Parameters
        ----------
        idt : int
            Id of the item.
        type_item : str
            Type of requested item. If empty, delete all items of the task.
        """
        path_task = self.xp_path / 'tasks' / '{:06}'.format(idt)
        if type_item == '':
            if path_task.exists():
                shutil.rmtree(str(path_task))
        else:
            if self._has_item(type_item, idt):
                filename = type_item + '.pickle'
                os.remove(str(path_task / filename))

    def _get_all_items(self, type_item, only_ids=False):
        """Get all items of the experiment.

        Parameters
        ----------
        type_item : str
            Type of requested item.
        only_ids : bool, optional
            Indicates if only ids of items are retrieved.

        Returns
        -------
        [dict, list]
            List of ids or dict keyed by the ids.
        """
        # create an empty container
        items = [] if only_ids else {}

        # get each item and insert it
        for task_path in (self.xp_path / 'tasks').glob('*'):
            filename = type_item + '.pickle'
            if (task_path / filename).exists():
                task_id = int(task_path.stem)
                if only_ids:
                    items.append(task_id)
                else:
                    items[task_id] = self._read_item(type_item, task_id)

        return items

    # Experiment management

    def _handle_exception_in_run_task(self, exc, idt, task_params):
        self.logger.exception(exc)
        exc_traceback = traceback.format_exception(*sys.exc_info())
        error_log = 'Task {}\n\nTask parameters:\n{}\n\n{}'.format(
            idt, str(task_params), ''.join(exc_traceback))
        self._write_item('error_log', idt, error_log)

    def add_tasks(self, data_params, problem_params, solver_params):
        """Add tasks using combinations of parameters.

        Parameters
        ----------
        data_params : dict
            Parameters of generation of the dataset.
        problem_params : dict
            Parameters of generation of the problem.
        solver_params : dict
            Parameters of generation of the solver.

        Raises
        ------
        TypeError
            If parameters are not dict.
        ValueError
            If parameters contain unexpected keys.

        Notes
        -----
        * All parameters should be dict, keyed by the name of the parameter and
          valued by a list.
        * Tasks are generated as the cartesian product of the sets of
          parameters. If parameters are added by successive calls to
          :meth:`~yafe.base.Experiment.add_tasks`, the resulting set of tasks
          is the cartesian product of all updated sets of parameters (which is
          equivalent to the set of tasks obtained by calling
          :meth:`~yafe.base.Experiment.add_tasks` only once with all the
          parameters).
        """
        # check that params are dict
        if not isinstance(data_params, dict):
            raise TypeError('data_params should be a dict')
        if not isinstance(problem_params, dict):
            raise TypeError('problem_params should be a dict')
        if not isinstance(solver_params, dict):
            raise TypeError('solver_params should be a dict')

        # convert non-iterable elements in list
        for key, value in data_params.items():
            if not hasattr(value, '__iter__') or isinstance(value, str):
                data_params[key] = [value]

        for key, value in problem_params.items():
            if not hasattr(value, '__iter__') or isinstance(value, str):
                problem_params[key] = [value]

        for key, value in solver_params.items():
            if not hasattr(value, '__iter__') or isinstance(value, str):
                solver_params[key] = [value]

        # update the _schema
        schema = self._schema

        def sort_key(a):
            return -1 if a is None else a

        for key, value in data_params.items():
            label = ('data', key)
            if label not in schema:
                errmsg = '{} is not a parameter for getting data.'
                raise ValueError(errmsg.format(key))
            else:
                schema[label] = sorted(
                    list(set(schema[label]).union(set(value))), key=sort_key)

        for key, value in problem_params.items():
            label = ('problem', key)
            if label not in schema:
                errmsg = '{} is not a parameter for getting the problem.'
                raise ValueError(errmsg.format(key))
            else:
                schema[label] = sorted(
                    list(set(schema[label]).union(set(value))), key=sort_key)

        for key, value in solver_params.items():
            label = ('solver', key)
            if label not in schema:
                errmsg = '{} is not a parameter for getting the solver.'
                raise ValueError(errmsg.format(key))
            else:
                schema[label] = sorted(
                    list(set(schema[label]).union(set(value))), key=sort_key)

        self._schema = schema

    def generate_tasks(self):
        """Generate the tasks from _schema.

        Already existing tasks are not rewritten.
        """

        # get _schema
        schema = self._schema

        # get the existing tasks
        existing_task_params = [
            dict(item)
            for item in self._get_all_items('task_params').values()]

        n_tasks_added = 0
        n_existing_tasks = len(existing_task_params)

        data_params = dict()
        problem_params = dict()
        solver_params = dict()

        for label in schema.keys():
            cat, key = label
            if cat == 'data':
                data_params[key] = schema[label]
            elif cat == 'problem':
                problem_params[key] = schema[label]
            else:  # cat == 'solver':
                solver_params[key] = schema[label]

        def param_grid(param_dict):
            res = []
            keys, values = zip(*param_dict.items())
            for val in itertools.product(*values):
                params = dict(zip(keys, val))
                res.append(params)
            return res

        # get cartesian products of parameters
        parameters = param_grid({'data_params':
                                 param_grid(data_params),
                                 'problem_params':
                                 param_grid(problem_params),
                                 'solver_params':
                                 param_grid(solver_params)})

        # for each set of parameter, write the corresponding task
        for params in [item for item in parameters
                       if item not in existing_task_params]:
            self._write_item('task_params',
                             n_existing_tasks + n_tasks_added,
                             params)
            n_tasks_added += 1

    def run_task_by_id(self, idt):
        """Run a task given an id.

        Parameters
        ----------
        idt : int
            Id of the task.
        """
        self.logger.info('Loading task %s', idt)
        task_params = self._read_item('task_params', idt)

        if task_params is None:
            err_msg = 'Unknown task: {}'.format(idt)
            self.logger.error(err_msg)
            raise ValueError(err_msg)

        self.logger.debug('Task %s: %s', idt, str(task_params))

        if self._has_item('error_log', idt):
            self.logger.info(
                'Task %s, remove error log from previous run', idt)
            self._delete_item(idt, 'error_log')

        try:
            if self._has_item('source_data', idt):
                self.logger.info(
                    'Task %s, source data already generated', idt)
                source_data = self._read_item('source_data', idt)
            else:
                self.logger.info('Task %s, generating source data', idt)
                source_data = self.get_data(**task_params['data_params'])

                self._write_item('source_data', idt, source_data)
            self.logger.debug('Task %s, source data : %s', idt,
                              {k: type(v) for k, v in source_data.items()})

        except Exception as exc:
            self._handle_exception_in_run_task(exc, idt, task_params)
            raise exc

        try:
            if self._has_item('problem_data', idt) \
                    and self._has_item('solution_data', idt):
                self.logger.info(
                    'Task %s, problem data and solution data already '
                    'generated', idt)
                problem_data = self._read_item('problem_data', idt)
                solution_data = self._read_item('solution_data', idt)
            else:
                self.logger.info('Task %s, loading problem', idt)
                problem = self.get_problem(**task_params['problem_params'])

                self.logger.info('Task %s, generating problem', idt)
                problem_params = inspect.signature(problem).parameters
                for param in source_data.keys():
                    if param not in problem_params:
                        msg = (
                            'Task {}, the output of get_data is not '
                            'compatible with the input of the problem '
                            'generated using get_problem. The parameter {} is '
                            'unexpected.'.format(idt, param))
                        raise TypeError(msg)
                problem_data, solution_data = problem(**source_data)

                self._write_item('problem_data', idt, problem_data)
                self._write_item('solution_data', idt, solution_data)

            self.logger.debug('Task %s, problem data : %s', idt,
                              {k: type(v) for k, v in problem_data.items()})
            self.logger.debug('Task %s, solution data : %s', idt,
                              {k: type(v) for k, v in solution_data.items()})
        except Exception as exc:
            self._handle_exception_in_run_task(exc, idt, task_params)
            raise exc

        try:
            if self._has_item('solved_data', idt):
                self.logger.info('Task %s, problem already solved', idt)
                solved_data = self._read_item('solved_data', idt)
            else:
                self.logger.info('Task %s, loading solver', idt)
                solver = self.get_solver(**task_params['solver_params'])

                self.logger.info('Task %s, solving problem', idt)
                solver_params = inspect.signature(solver).parameters
                for param in problem_data.keys():
                    if param not in solver_params:
                        msg = (
                            'Task {}, the output of the problem generated '
                            'using get_problem is not compatible with the '
                            'input of the solver generated using get_solver. '
                            'The parameter {} is unexpected.'.format(idt,
                                                                     param))
                        raise TypeError(msg)
                solved_data = solver(**problem_data)

                self._write_item('solved_data', idt, solved_data)

            self.logger.debug('Task %s, solver data : %s', idt,
                              {k: type(v) for k, v in solved_data.items()})
        except Exception as exc:
            self._handle_exception_in_run_task(exc, idt, task_params)
            raise exc

        try:
            if self._has_item('result', idt):
                self.logger.info('Task %s, performance already measured', idt)
                result = self._read_item('result', idt)
            else:
                self.logger.info('Task %s, measuring performance', idt)
                result = self.measure(
                    task_params=task_params,
                    source_data=source_data,
                    problem_data=problem_data,
                    solution_data=solution_data,
                    solved_data=solved_data)
                self._write_item('result', idt, result)

            self.logger.debug('Task %s, result data: %s', idt,
                              {k: type(v) for k, v in result.items()})
        except Exception as exc:
            self._handle_exception_in_run_task(exc, idt, task_params)
            raise exc

    def get_pending_task_ids(self):
        """
        Get tasks that have not been run and successfully completed.

        Returns
        -------
        set
            Set of pending task ids
        """

        # get the list of task to do
        no_tasks = self._get_all_items('task_params', only_ids=True)
        no_results = self._get_all_items('result', only_ids=True)

        return set(no_tasks).difference(no_results)

    def launch_experiment(self, task_ids=None):
        """Launch an experiment.

        Parameters
        ----------
        task_ids : str or list of ints or None
            List of task ids to run. If str, task ids must be separated by a
            comma. If None, run pending tasks only to avoid running the same
            tasks several times.
        """
        if task_ids is None:
            task_ids = self.get_pending_task_ids()
        elif isinstance(task_ids, str):
            task_ids = [int(task) for task in task_ids.split(',')]

        self.logger.info('Launching {} tasks'.format(len(task_ids)))

        for idt in task_ids:
            try:
                self.run_task_by_id(idt)
            except Exception as exc:
                pass

    def collect_results(self):
        """Collect results from an experiment.

        Merge all the task results in a :class:`numpy.ndarray` and save the
        resulting data with axes labels and values in a file.

        The resulting data are stored in the data directory of the experiment
        in a file named ``results.npz``. This file contains three
        :class:`numpy.ndarray`:

          - results: the collected results,
          - axes_labels: the labels of the axes of results,
          - axes_values: the values corresponding to the indexes for all
            axes of results.

        The dimensions of the main array ``results`` match the parameters of
        the tasks and the performance measures (last dimension). More
        precisely, ``results[i_0, ..., i_j, ..., i_N]`` is the value of the
        performance measure axes_values[N][i_N] for parameter values
        ``axes_values[j][i_j]``, ``j < N``, related to parameter names
        ``axes_label[j]``.

        The stored data can be loaded using the method
        :meth:`~yafe.base.Experiment.load_results`.

        Raises
        ------
        RuntimeError
            If no results have been computed prior to running the method.
        """
        # get the tasks and the results
        tasks_params = self._get_all_items('task_params')
        results = self._get_all_items('result')

        if results:
            measures_names = list(next(iter(results.values())).keys())
        else:
            raise RuntimeError('No result available. Results must be computed '
                               'before running collect_results().')

        self.logger.info('Collecting {} tasks.'.format(len(tasks_params)))

        # add axes corresponding to the parameters of the experiment
        schema = self._schema

        shape = [len(value) for value in schema.values()]
        shape.append(len(measures_names))
        dtype = np.float64
        array_results = np.empty(shape, dtype=dtype)
        array_results.fill(np.nan)

        # load the results and store them
        missing_tasks = dict()
        for measure in measures_names:
            missing_tasks[measure] = list()

        for idt, task in tasks_params.items():

            indices = []
            for (prefix, name), ticks in schema.items():
                indices.append(ticks.index(task[prefix + '_params'][name]))
            indices.append(slice(None))
            indices = tuple(indices)

            result = np.empty(len(measures_names))
            result.fill(np.nan)
            for idm, measure in enumerate(measures_names):
                try:
                    result[idm] = results[idt][measure]
                except KeyError:
                    missing_tasks[measure].append(idt)

            array_results[indices] = result

        for measure in measures_names:
            self.logger.info('Measure `{}`: {} missing tasks.'.format(
                measure, len(missing_tasks[measure])))
            self.logger.debug(missing_tasks)

        axes_labels = [cat + '_' + param for (cat, param) in schema.keys()]
        axes_labels.append('measure')
        axes_labels = np.array(axes_labels, dtype=object)
        axes_values = np.empty_like(axes_labels, dtype=object)
        for ind, val in enumerate(schema.values()):
            axes_values[ind] = np.array(val)
        axes_values[-1] = np.array(measures_names, dtype=object)

        np.savez_compressed(
            str(self.xp_path / 'results.npz'), results=array_results,
            axes_labels=axes_labels, axes_values=axes_values)

    def load_results(self, array_type='numpy'):
        """Load the collected results from an experiment.

        Load the data stored by the method
        :meth:`~yafe.base.Experiment.collect_results` in the file
        ``results.npz``.

        Parameters
        ----------
        array_type : {'numpy', 'xarray'}
            Type of output. With the option ``'numpy'``, the output is split
            into three :class:`numpy.ndarray`, while with the option
            ``'xarray'`` the output is a single :class:`xarray.DataArray`.

        Returns
        -------
        (results, axes_labels, axes_values) : tuple of :class:`numpy.ndarray`
            Returned when using the default option ``array_type='numpy'``.
            See :meth:`~yafe.base.Experiment.collect_results` for a detailed
            description.
        xresults : :class:`xarray.DataArray`
            Returned when using the option ``array_type='xarray'``. This is
            equivalent to the case ``array_type='numpy'``, combining all
            three arrays into a single data structure for improved consistency
            and ease of use. ``xresults`` can be obtained from
            ``(results, axes_labels, axes_values)`` as ::

                xresults = xarray.DataArray(data=results,
                                            dims=axes_labels,
                                            coords=axes_values)

        Raises
        ------
        ImportError
            If the xarray package is not installed and the
            ``array_type='xarray'`` option is used.
        ValueError
            If an unrecognized value is given for ``array_type``.
        RuntimeError
            If no collected results file can be found.
        """
        if (self.xp_path / 'results.npz').exists():
            data = np.load(str(self.xp_path / 'results.npz'),
                           allow_pickle=True)
            if array_type == 'numpy':
                return (data['results'],
                        data['axes_labels'],
                        data['axes_values'])
            elif array_type == 'xarray':
                try:
                    import xarray
                except ImportError:
                    raise ImportError('Please install the xarray package to '
                                      'run the load_results() method with the '
                                      'array_type="xarray" option.')
                return xarray.DataArray(data=data['results'],
                                        dims=data['axes_labels'],
                                        coords=data['axes_values'])
            else:
                errmsg = ('Unrecognized value for array_type, the valid '
                          'options are "numpy" and "xarray".')
                raise ValueError(errmsg)
        else:
            raise RuntimeError('No file with collected results available. '
                               'Results must be computed and collected before '
                               'running load_results().')

    def _get_status(self, summary):
        """Return a string describing the satus of the tasks.

        Parameters
        ----------
        summary : bool
            Indicates if only a summary of the status is provided or not.

        Returns
        -------
        str
            Status of the tasks.
        """
        ids_tasks = self._get_all_items('task_params', only_ids=True)
        ids_results = self._get_all_items('result', only_ids=True)
        ids_errors = self._get_all_items('error_log', only_ids=True)

        status = ''

        if summary:
            status += 'Number of tasks: {}\n'.format(len(ids_tasks))
            status += 'Number of finished tasks: {}\n'.format(len(ids_results))
            status += 'Number of failed tasks: {}\n'.format(len(ids_errors))
            status += 'Number of pending tasks: {}\n'.format(
                len(ids_tasks) - len(ids_results))
        else:
            for id_task in sorted(ids_tasks):
                width = 26
                nb_digits_id = 6
                id_format = '{{:0{nb}d}}'.format(nb=nb_digits_id)
                status += id_format.format(id_task)

                if id_task in ids_results:
                    msg = 'OK'
                else:
                    if id_task in ids_errors:
                        msg = 'ERROR'
                    else:
                        msg = 'NOT CALCULATED'
                status += '.' * (width-nb_digits_id-len(msg)) + msg + '\n'
        return status

    def display_status(self, summary=True):
        """Display the status of the tasks.

        Parameters
        ----------
        summary : bool, optional
            Indicates if only a summary of the status is displayed or not.
        """
        print(self._get_status(summary))

    def get_task_data_by_id(self, idt):
        """Get data for a task specified by its id.

        Parameters
        ----------
        idt : int
            Id of the task.

        Returns
        -------
        dict
            Task data.

        Raises
        ------
        ValueError
            If ``idt`` is not a valid value.
        """
        task_params = self._read_item('task_params', idt)
        if task_params is None:
            err_msg = 'Unknown task: {}'.format(idt)
            self.logger.error(err_msg)
            raise ValueError(err_msg)
        source_data = self._read_item('source_data', idt)
        problem_data = self._read_item('problem_data', idt)
        solution_data = self._read_item('solution_data', idt)
        solved_data = self._read_item('solved_data', idt)
        result = self._read_item('result', idt)
        return {'id_task': idt,
                'task_params': task_params,
                'source_data': source_data,
                'problem_data': problem_data,
                'solution_data': solution_data,
                'solved_data': solved_data,
                'result': result}

    def get_task_data_by_params(self, data_params, problem_params,
                                solver_params):
        """Get data for a task specified by its parameters.

        Parameters
        ----------
        data_params : dict
            Parameters of generation of the dataset.
        problem_params : dict
            Parameters of generation of the problem.
        solver_params : dict
            Parameters of generation of the solver.

        Returns
        -------
        dict
            Task data.

        Raises
        ------
        TypeError
            If parameters are not dict.
        ValueError
            If parameters contain unexpected keys or unexpected values.
        """
        if not isinstance(data_params, dict) \
                or not isinstance(problem_params, dict) \
                or not isinstance(solver_params, dict):
            msg = 'Parameters must be of type dict.'
            raise TypeError(msg)

        msg = 'Wrong keys in {}, the expected keys are: {}.'
        data_params_names = inspect.signature(self.get_data).parameters
        if set(data_params.keys()) != set(data_params_names):
            raise ValueError(
                msg.format('data_params', ', '.join(data_params_names)))

        problem_params_names = inspect.signature(self.get_problem).parameters
        if set(problem_params.keys()) != set(problem_params_names):
            raise ValueError(
                msg.format('problem_params', ', '.join(problem_params_names)))

        solver_params_names = inspect.signature(self.get_solver).parameters
        if set(solver_params.keys()) != set(solver_params_names):
            raise ValueError(
                msg.format('solver_params', ', '.join(solver_params_names)))

        id_task = None
        all_tasks_params = self._get_all_items('task_params')
        params = {'data_params': data_params,
                  'problem_params': problem_params,
                  'solver_params': solver_params}
        for idt, task_params in all_tasks_params.items():
            if params == task_params:
                id_task = idt
                break

        if id_task is not None:
            return self.get_task_data_by_id(id_task)
        else:
            err_msg = 'No task exists for the given parameters.'
            self.logger.error(err_msg)
            raise ValueError(err_msg)



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# yafe: Yet Another Framework for Experiments.
#
# Licence
# -------
# This file is part of yafe.
#
# yafe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Utils classes and functions for yafe.

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
.. moduleauthor:: Florent Jaillet
"""
import configparser
import importlib
import logging
import os
from pathlib import Path
import stat
import sys


class ConfigParser(configparser.ConfigParser):
    """Configuration file parser for yafe.

    This class inherits from ConfigParser in the configparser module.

    It enables reading the yafe configuration file ``$HOME/.config/yafe.conf``
    at initialization of an experiment. It also provides a method to properly
    read a path in the configuration file, and a static method to generate a
    basic empty configuration file.
    """
    _config_path = Path(os.path.expanduser('~')) / '.config' / 'yafe.conf'

    def __init__(self):
        super().__init__()
        loaded_files = super().read(self._config_path)
        if not len(loaded_files):
            print('Generation of the config file.')
            ConfigParser.generate_config()
            print('Please update config file at {}'.format(self._config_path))
            super().read(self._config_path)

    def get_path(self, section, option, check_exists=True):
        """Get the path filled in a given option of a given section.

        Parameters
        ----------
        section : str
            Name of the section.
        option : str
            Name of the option.
        check_exists : bool, optional
            Indicates if the existence of the path is checked.

        Returns
        -------
        pathlib.Path or None
            Path if the option is defined, None otherwise.

        Raises
        ------
        IOError
            If the parameter ``check_exists`` is set to ``True`` and the path
            does not exist.
        """
        path = super().get(section, option)

        if path == '':
            return None

        path = Path(path)

        if check_exists and not path.exists():
            errmsg = 'Path {} does not exist.'
            raise IOError(errmsg.format(path))

        return path

    @staticmethod
    def generate_config():
        """Generate an empty configuration file.

        The generated configuration file is stored in
        ``$HOME/.config/yafe.conf``.
        """

        config = configparser.ConfigParser(allow_no_value=True)

        config.add_section('USER')
        config.set('USER', '# user directory (data generated by experiments)')
        config.set('USER', 'data_path', '')

        config.add_section('LOGGER')
        config.set('LOGGER', '# path to the log file')
        config.set('LOGGER', 'path', '')

        with open(str(ConfigParser._config_path), 'w') as config_file:
            config.write(config_file)


def get_logger(name='', to_file=True, to_console=True, logger_path=None):
    """Return a yafe logger with the given name.

    Parameters
    ----------
    name : str
        Name of the logger.
    to_file : bool
        Indicates if the logger writes processing and debug information in a
        file.
    to_console : bool
        Indicates if the logger displays processing information in the console.
    logger_path : None or str or pathlib.Path
        Full file path where the data must be written by the logger when
        ``to_file == True``.
        If ``logger_path`` is ``None`` the logger path given in the yafe
        configuration file is used.

    Returns
    -------
    logging.Logger
        A logger with the given name prefixed by ``yafe.``.

    Notes
    -----
    The name of the logger is automatically prefixed by ``yafe.``.
    """
    logger = logging.getLogger('yafe.{}'.format(name))
    logger.handlers = []
    logger.setLevel(logging.DEBUG)

    if to_file:
        if logger_path is None:
            config = ConfigParser()
            logger_path = config.get_path('LOGGER', 'PATH', check_exists=False)
        else:
            logger_path = Path(logger_path)
        if not logger_path.exists():
            logger_path.touch()
        file_handler = logging.FileHandler(str(logger_path))
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(
            logging.Formatter(
                '%(asctime)s - %(name)s - %(levelname)s - %(message)s'))

        logger.addHandler(file_handler)

    if to_console:
        console_handler = logging.StreamHandler(sys.stdout)
        console_handler.setLevel(logging.INFO)
        console_handler.setFormatter(
            logging.Formatter(
                '%(asctime)s - %(name)s - %(levelname)s - %(message)s'))

        logger.addHandler(console_handler)

    return logger


def generate_oar_script(script_file_path, xp_var_name, task_ids=None,
                        batch_size=1, oar_walltime='02:00:00',
                        activate_env_command=None, use_gpu=False):
    """Generate a script to launch an experiment using OAR.

    Tasks are divided into batches that are executed by oar jobs.

    The resulting script is written in the experiment folder, and the command
    to launch the jobs with OAR is displayed in the terminal.

    An example script illustrating how to use
    :func:`yafe.utils.generate_oar_script` is available in the corresponding
    :ref:`tutorial <tutorial_oar>`.

    Parameters
    ----------
    script_file_path : str
        File path to the script that defines the experiment.
    xp_var_name : str
        Name of the variable containing the experiment in the script.
    task_ids : list
        List of tasks ids to run.
        If ``task_ids`` is ``None``, the list of pending tasks of the
        experiment is used.
    batch_size : int
        Number of tasks run in each batch.
    oar_walltime : str
        Wall time for each OAR job ('HH:MM:SS').
    activate_env_command : str or None
        Optional command that must be run to activate a Python virtual
        environment before launching the experiment.
        Typically, this is a command of the form
        ``source some_virtual_env/bin/activate`` when using virtualenv and
        ``source activate some_conda_env`` when using conda.
        If ``activate_env_command`` is ``None``, no virtual environment is
        activated.
    use_gpu : bool
        Flag specifying if a gpu ressource is needed when running the
        experiment.
    """
    script_file_path = Path(script_file_path)
    script_dir = script_file_path.parent
    script_name = script_file_path.stem

    sys.path.append(str(script_dir))
    mod = importlib.import_module(script_name)
    xp = getattr(mod, xp_var_name)

    if task_ids is None:
        task_ids = xp.get_pending_task_ids()

    # split and save the tasks
    task_ids = list(map(str, task_ids))
    batches = [
        task_ids[i:(i + batch_size)]
        for i in range(0, len(task_ids), batch_size)
    ]
    file_path = xp.xp_path / 'listoftasks.txt'

    with open(str(file_path), 'wt') as fout:
        fout.write('\n'.join(map(lambda batch: ','.join(batch), batches)))

    # generate and save script
    script_path = Path(os.path.abspath(script_file_path))
    script_dir = script_path.parent
    script_name = script_path.stem

    script = '#!/bin/sh\n'

    # define the OAR parameters
    script += '#OAR -n {}\n'.format(xp.name)
    script += '#OAR --array-param-file {}\n'.format(str(file_path))
    script += '#OAR -O {}/%jobid%.out\n'.format(xp.xp_path)
    script += '#OAR -E {}/%jobid%.err\n'.format(xp.xp_path)
    script += '#OAR -l walltime={}\n'.format(oar_walltime)
    if use_gpu:
        script += '#OAR -p gpu IS NOT NULL\n'
    else:
        script += '#OAR -p gpu IS NULL\n'

    script += 'echo "OAR_JOB_ID: $OAR_JOB_ID"\n'
    script += 'echo "OAR_ARRAY_ID: $OAR_ARRAY_ID"\n'
    script += 'echo "OAR_ARRAY_INDEX: $OAR_ARRAY_INDEX"\n'

    # activate the virtual env
    if activate_env_command is not None and len(activate_env_command) > 0:
        script += '{}\n'.format(activate_env_command)

    # python command
    script += 'echo "Running {}.launch_experiment(task_ids=\'$1\')"\n'.format(
        xp_var_name)
    script += 'python -c "import sys; sys.path.append(\'{0}\'); ' \
        'from {1} import {2}; ' \
        '{2}.launch_experiment(task_ids=\'$1\')"\n'.format(
            script_dir, script_name, xp_var_name)
    script += 'exit $?'

    script_path = xp.xp_path / 'script_oar.sh'
    with script_path.open('w') as file:
        file.write(script)
    status = os.stat(script_path)
    os.chmod(script_path, status.st_mode | stat.S_IXUSR)

    print('oarsub -S {}'.format(str(script_path)))



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# yafe: Yet Another Framework for Experiments.
#
# Licence
# -------
# This file is part of yafe.
#
# yafe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Example experiment to be deployed on a cluster grid with OAR.

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
.. moduleauthor:: Florent Jaillet

This example and the following workflow illustrates the design of a ``yafe``
experiment, its execution on a cluster and how to collect and exploit
results remotely or locally.

* On the cluster frontend, run this script to generate the experiment,
  its tasks, and the OAR script (see ``__main__`` below). Note that if you are
  using a virtual environment, you should add it as an argument when calling
  function ``generate_oar_script`` and activate it before running this script.

* Run the tasks on the cluster by executing the command line displayed by
  the script.

* Optionally, collect and exploit results on the cluster by opening a python
  session and by running::

      from example_oar import experiment
      experiment.collect_results()
      results, axes_labels, axes_values = experiment.load_results()

* In order to exploit results on a local machine, download the
  experiment directory and all its contents into your local data path and
  run the above code locally. Note that downloading all the directory
  is needed to preserve the consistency of the task indexing, especially if
  you want to examine some specific tasks or add new ones.

* If you only want to exploit results without downloading all the
  experiment data stored in the directory, you should collect the
  results remotely on the cluster and download files ``_schema.pickle`` and
  ``results.npz`` into your local data directory. You can then exploit the
  results using ::

      from example_oar import experiment
      results, axes_labels, axes_values = experiment.load_results()

  Be aware that only the method ``load_results`` will be usable and that the
  ``experiment`` object will be in an inconsistent state for all the other
  methods.
"""
import numpy as np

from yafe import Experiment
from yafe.utils import generate_oar_script


def get_sine_data(f0, signal_len=1000):
    return {'signal': np.sin(2*np.pi*f0*np.arange(signal_len))}


class SimpleDenoisingProblem:
    def __init__(self, snr_db):
        self.snr_db = snr_db

    def __call__(self, signal):
        random_state = np.random.RandomState(0)
        noise = random_state.randn(*signal.shape)
        observation = signal + 10 ** (-self.snr_db / 20) * noise \
            / np.linalg.norm(noise) * np.linalg.norm(signal)
        problem_data = {'observation': observation}
        solution_data = {'signal': signal}
        return problem_data, solution_data

    def __str__(self):
        return 'SimpleDenoisingProblem(snr_db={})'.format(self.snr_db)


class SmoothingSolver:
    def __init__(self, filter_len):
        self.filter_len = filter_len

    def __call__(self, observation):
        smoothing_filter = np.hamming(self.filter_len)
        smoothing_filter /= np.sum(smoothing_filter)
        return {'reconstruction': np.convolve(observation, smoothing_filter,
                                              mode='same')}

    def __str__(self):
        return 'SmoothingSolver(filter_len={})'.format(self.filter_len)


def measure(solution_data, solved_data, task_params=None, source_data=None,
            problem_data=None):
    euclidian_distance = np.linalg.norm(solution_data['signal']
                                        - solved_data['reconstruction'])
    sdr = 20 * np.log10(np.linalg.norm(solution_data['signal'])
                        / euclidian_distance)
    inf_distance = np.linalg.norm(solution_data['signal']
                                  - solved_data['reconstruction'], ord=np.inf)
    return {'sdr': sdr,
            'euclidian_distance': euclidian_distance,
            'inf_distance': inf_distance}


def create_tasks(exp):
    data_params = {'f0': np.arange(0.01, 0.1, 0.01), 'signal_len': [1000]}
    problem_params = {'snr_db': [-10, 0, 30]}
    solver_params = {'filter_len': 2**np.arange(6, step=2)}
    exp.add_tasks(data_params=data_params,
                  problem_params=problem_params,
                  solver_params=solver_params)
    exp.generate_tasks()


experiment = Experiment(name='yafe_oar_example',
                        get_data=get_sine_data,
                        get_problem=SimpleDenoisingProblem,
                        get_solver=SmoothingSolver,
                        measure=measure)


if __name__ == '__main__':
    create_tasks(experiment)
    generate_oar_script(script_file_path=__file__,
                        xp_var_name='experiment',
                        batch_size=4,
                        oar_walltime='00:05:00')



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# yafe: Yet Another Framework for Experiments.
#
# Licence
# -------
# This file is part of yafe.
#
# yafe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Test of the module :module:`yafe.base`

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
.. moduleauthor:: Florent Jaillet
"""
import contextlib
import json
import os
from pathlib import Path
from itertools import product
from shutil import rmtree
from io import StringIO
import sys
import tempfile
import unittest
from unittest.mock import patch
import warnings

import numpy as np

from yafe.base import Experiment
from yafe.utils import ConfigParser


def create_temp_test_config():
    temp_dir = tempfile.mkdtemp(prefix='yafe_')
    ConfigParser._config_path = temp_dir / Path('yafe.conf')
    config_parser = ConfigParser()
    data_path = temp_dir / Path('data')
    data_path.mkdir()
    config_parser.set('USER', 'data_path', str(data_path))
    logger_path = temp_dir / Path('yafe.log')
    config_parser.set('LOGGER', 'path', str(logger_path))
    with open(str(ConfigParser._config_path), 'w') as config_file:
        config_parser.write(config_file)
    return temp_dir


def get_data_path():
    config = ConfigParser()
    data_path = config.get_path('USER', 'data_path')
    return data_path


# Mock component with arbitrary behavior
class MockSolver:
    def __init__(self, f, seed):
        self.f = f
        self.seed = seed

    def __call__(self, X):
        if self.f == 1000:
            raise ValueError('Exception in solve for testing')
        return {'X': X + self.f}


class MockProblem:
    def __init__(self, c, d, e, seed):
        self.value = c + d + len(e)
        self.seed = seed

    def __call__(self, X):
        np.random.seed(self.seed)
        if self.value == 1000:
            raise ValueError('Exception in generate for testing')
        problem_data = {'X': X + self.value}
        solution_data = {'X': X + self.value + 1}
        return problem_data, solution_data


def measure(task_params, source_data, problem_data, solution_data,
            solved_data):
    if task_params['data_params']['a'] == 10000:
        raise ValueError('Exception in measure for testing')
    return {'snr': 0.2, 'error': 1.6}


class MockExperiment(Experiment):
    def __init__(self, name, force_reset=False, data_path=None):

        def get_data(a, b, seed):
            if a == 1000:
                raise ValueError('Exception in get_data for testing')
            np.random.seed(seed)
            return {'X': np.arange(a + b)}

        def get_problem(c, d, e, seed):
            return MockProblem(c, d, e, seed)

        def get_solver(f, seed):
            return MockSolver(f, seed)

        super().__init__(name, get_data=get_data, get_problem=get_problem,
                         get_solver=get_solver, measure=measure,
                         force_reset=force_reset, data_path=data_path)


# Mock components to track the tasks contents
class MockTrackData:
    def __call__(self, data_param1):
        return {'mock_data': np.array([data_param1])}

    @staticmethod
    def get_params():
        return {
            'data_param1': [-1, 2, -4],  # not sorted
        }

    @staticmethod
    def get_additional_params():
        return {'data_param1': [6]}


class MockTrackProblem:
    # argument names not in lexicographic order
    def __init__(self, problem_param3, problem_param1, problem_param2):
        self.problem_param3 = problem_param3
        self.problem_param1 = problem_param1
        self.problem_param2 = problem_param2

    def __call__(self, mock_data):
        mock_pb_data = np.concatenate((mock_data, [self.problem_param3,
                                                   self.problem_param1,
                                                   self.problem_param2]))
        problem_data = {'mock_pb_data': mock_pb_data}
        mock_sol_data = np.concatenate((mock_data, [self.problem_param3,
                                                    self.problem_param1,
                                                    self.problem_param2]))
        solution_data = {'mock_sol_data': mock_sol_data}
        return problem_data, solution_data

    @staticmethod
    def get_params():
        return {
            'problem_param3': [10, 11, 12],  # sorted
            'problem_param1': [20],  # singleton
            'problem_param2': [33, 32],  # not sorted
        }

    @staticmethod
    def get_additional_params():
        return {'problem_param2': [20]}


class MockTrackSolver:
    # order of arguments solver_param2 and solver_param1 differ from order in
    # dict
    def __init__(self, solver_param2, solver_param1):
        self.solver_param2 = solver_param2
        self.solver_param1 = solver_param1

    def __call__(self, mock_pb_data):
        mock_solver_data = np.concatenate(
            (mock_pb_data, [self.solver_param2, self.solver_param1]))
        return {'mock_solver_data': mock_solver_data}

    @staticmethod
    def get_params():
        return {
            'solver_param1': [41],  # singleton
            'solver_param2': [54, 53],  # not sorted
        }

    @staticmethod
    def get_additional_params():
        return{
            'solver_param1': [39],
            'solver_param2': [54, 53.5],
        }


def mock_track_measure(solved_data, source_data=None, task_params=None,
                       problem_data=None, solution_data=None):
    return {
        'meas_data_param1': solved_data['mock_solver_data'][0],
        'meas_problem_param3': solved_data['mock_solver_data'][1],
        'meas_problem_param1': solved_data['mock_solver_data'][2],
        'meas_problem_param2': solved_data['mock_solver_data'][3],
        'meas_solver_param2': solved_data['mock_solver_data'][4],
        'meas_solver_param1': solved_data['mock_solver_data'][5],
    }


def check_tracked_results(results, axes_labels, axes_values):
    # Check shape
    np.testing.assert_array_equal(results.shape,
                                  [a.size for a in axes_values])

    # Check axes_labels
    measure_names = axes_values[axes_labels == 'measure'][0]
    for i, name in enumerate(measure_names):
        assert name.startswith('meas_')
        if axes_labels[i].startswith('data_'):
            assert name[len('meas_'):] == axes_labels[i][len('data_'):]
        elif axes_labels[i].startswith('problem_'):
            assert name[len('meas_'):] == axes_labels[i][len('problem_'):]
        elif axes_labels[i].startswith('solver_'):
            assert name[len('meas_'):] == axes_labels[i][len('solver_'):]
        else:
            raise ValueError(
                '{} does not begin with {}.'.format(
                    axes_labels[i], ('data_', 'problem_', 'solver_')))

    # Check first element in results
    value = results[(0,) * (results.ndim - 1)]
    for i in range(results.ndim-1):
        assert value[i] == axes_values[i][0]

    # Check all elements in results
    axes_values_params = axes_values[axes_labels != 'measure']
    indexes = [np.arange(a.size) for a in axes_values_params]
    for multi_index, multi_value in zip(product(*indexes),
                                        product(*axes_values_params)):
        for j in range(len(multi_value)):
            assert results[multi_index][j] == axes_values[j][multi_index[j]]


class TestExperiment(unittest.TestCase):

    def setUp(self):
        self.item = {'a': 1, 'e': 'alpha', 'f': [0, 2, 8], 'g': (4, 7)}

        self.data_params = {'a': [1, 2, 3], 'b': [4, 5], 'seed': 12}
        self.problem_params = {'c': [3, 2, 1],
                               'd': 5, 'e': ['alpha', 'beta'], 'seed': 12}
        self.solver_params = {'f': 31, 'seed': 12}

        all_values = list(self.data_params.values()) \
            + list(self.problem_params.values()) \
            + list(self.solver_params.values())
        all_len = [len(val) for val in all_values if isinstance(val, list)]
        self.expected_nb_tasks = np.product(all_len)

        self._temp_dir = create_temp_test_config()

    def tearDown(self):
        rmtree(self._temp_dir)

    def test_init(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:

            xp_name = Path(tmpdir).stem
            Path(tmpdir).rmdir()

            # Init a new experiment
            xp = MockExperiment(xp_name)

            self.assertEqual(str(xp.xp_path), tmpdir, msg='xp path')
            self.assertTrue(xp.xp_path.exists(), msg='existence of xp folder')
            self.assertTrue((xp.xp_path / 'tasks').exists(),
                            msg='existence of tasks folder')

            # add files in subfolders
            (xp.xp_path / 't.test').touch()
            (xp.xp_path / 'tasks' / 'tas.test').touch()

            # load existing xp
            xp = MockExperiment(xp_name)

            self.assertTrue((xp.xp_path / 't.test').exists(),
                            msg='file exists in xp folder ')
            self.assertTrue((xp.xp_path / 'tasks' / 'tas.test').exists(),
                            msg='file exists in tasks folder')

            # load existing xp without _schema
            os.remove(xp.xp_path / '_schema.pickle')
            xp = MockExperiment(xp_name)

            self.assertFalse((xp.xp_path / 't.test').exists(),
                             msg='file does not exists in xp folder')
            self.assertFalse((xp.xp_path / 'tasks' / 'tas.test').exists(),
                             msg='file does not exist in tasks folder')
            self.assertEqual(xp._schema,
                             {('data', 'a'): [],
                              ('data', 'b'): [],
                              ('data', 'seed'): [],
                              ('problem', 'c'): [],
                              ('problem', 'd'): [],
                              ('problem', 'e'): [],
                              ('problem', 'seed'): [],
                              ('solver', 'f'): [],
                              ('solver', 'seed'): []},
                             msg='_schema')

            # add files in subfolders
            (xp.xp_path / 't.test').touch()
            (xp.xp_path / 'tasks' / 'tas.test').touch()

            # check that the force_reset works
            xp = MockExperiment(xp_name, force_reset=True)

            self.assertFalse((xp.xp_path / 't.test').exists(),
                             msg='file does exists in xp folder ')
            self.assertFalse((xp.xp_path / 'tasks' / 'tas.test').exists(),
                             msg='file does exists in tasks folder')
            self.assertEqual(xp._schema,
                             {('data', 'a'): [],
                              ('data', 'b'): [],
                              ('data', 'seed'): [],
                              ('problem', 'c'): [],
                              ('problem', 'd'): [],
                              ('problem', 'e'): [],
                              ('problem', 'seed'): [],
                              ('solver', 'f'): [],
                              ('solver', 'seed'): []},
                             msg='_schema')

            # load existing xp with incoherent _schema.pickle file
            def get_data(a, b):
                return {'X': np.arange(a + b)}
            error_msg = ('The schema of the Experiment does not match the '
                         'schema of the Experiment with the same name stored '
                         'in ')
            with self.assertRaisesRegex(RuntimeError, error_msg):
                xp = Experiment(
                    xp_name, get_data=get_data, get_problem=MockProblem,
                    get_solver=MockSolver, measure=measure)

            os.remove(xp.xp_path / '_schema.pickle')
            with self.assertRaises(FileNotFoundError, msg='Missing _schema'):
                xp._schema

        tempfile.TemporaryDirectory(dir=str(get_data_path()))

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem

            def measure_wrong_inputs(task_params, problem_data, solved_data):
                return {'snr': 0.2, 'error': 1.6}

            def get_data(a, b, seed):
                return {'X': np.arange(a + b)}

            error_msg = ('Wrong signature for measure. measure must accept '
                         'the following parameters: task_params, source_data, '
                         'problem_data, solution_data, solved_data.')
            with self.assertRaisesRegex(TypeError, error_msg):
                _ = Experiment(name=xp_name,
                               get_data=get_data,
                               get_problem=MockProblem,
                               get_solver=MockSolver,
                               measure=measure_wrong_inputs)

    def test_str(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)
            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()

            str_xp = str(xp)

            expected_start = (
                'Name: tmp.*\n'
                'Data stored in: /.*\n'
                'Debug information stored in: /.*\n'
                'Data provider: <function MockExperiment.__init__.<locals>.'
                'get_data at 0x[0-9a-f]*>\n'
                'Problem provider : <function MockExperiment.__init__.'
                '<locals>.get_problem at 0x[0-9a-f]*>\n'
                'Solver provider: <function MockExperiment.__init__.'
                '<locals>.get_solver at 0x[0-9a-f]*>\n'
                'Measure: <function measure at 0x[0-9a-f]*>\n')
            self.assertRegex(str_xp, expected_start)

            expected_end = (
                '\n'
                'Parameters:\n'
                ' Data parameters:\n'
                '  a: [1, 2, 3]\n'
                '  b: [4, 5]\n'
                '  seed: [12]\n'
                ' Problem parameters:\n'
                '  c: [1, 2, 3]\n'
                '  d: [5]\n'
                "  e: ['alpha', 'beta']\n"
                '  seed: [12]\n'
                ' Solver parameters:\n'
                '  f: [31]\n'
                '  seed: [12]\n'
                '\n'
                'Status:\n'
                'Number of tasks: 36\n'
                'Number of finished tasks: 0\n'
                'Number of failed tasks: 0\n'
                'Number of pending tasks: 36\n')
            self.assertEqual(str_xp[-len(expected_end):], expected_end)

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            def get_data(a, b, seed):
                np.random.seed(seed)
                return {'X': np.arange(a + b)}
            xp_name = Path(tmpdir).stem
            xp = Experiment(xp_name,
                            get_data=get_data,
                            get_problem=MockProblem,
                            get_solver=MockSolver,
                            measure=measure,
                            log_to_file=False,
                            log_to_console=True)
            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()

            str_xp = str(xp)

            expected_start = (
                'Name: tmp.*\n'
                'Data stored in: /.*\n'
                'Not storing any debug information\n'
                'Data provider: <function TestExperiment.test_str.<locals>.'
                'get_data at 0x[0-9a-f]*>\n'
                'Problem provider : '
                "<class 'yafe.tests.test_base.MockProblem'>\n"
                "Solver provider: <class 'yafe.tests.test_base.MockSolver'>\n"
                'Measure: <function measure at 0x[0-9a-f]*>\n')
            self.assertRegex(str_xp, expected_start)

    def test_reset(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            Path(tmpdir).rmdir()

            # Init a new experiment
            xp = MockExperiment(xp_name)

            # add files in subfolders
            (xp.xp_path / 't.test').touch()
            (xp.xp_path / 'tasks' / 'tas.test').touch()

            # Reset the xp
            xp.reset()

            self.assertFalse((xp.xp_path / 't.test').exists(),
                             msg='file does not exist in xp folder')
            self.assertFalse((xp.xp_path / 'tasks' / 'tas.test').exists(),
                             msg='file does not exist in tasks folder')

            # The following is needed to reach 100% coverage
            rmtree(str(xp.xp_path))
            xp.reset()

            self.assertFalse((xp.xp_path / 't.test').exists(),
                             msg='file does not exist in xp folder')
            self.assertFalse((xp.xp_path / 'tasks' / 'tas.test').exists(),
                             msg='file does not exist in tasks folder')

    def test_write_read_delete_item(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp._write_item('name_item', 10, self.item)

            self.assertTrue(
                (Path(tmpdir) / 'tasks' / '000010').exists(),
                msg='existence of folder')

            self.assertTrue(
                (Path(tmpdir) / 'tasks' / '000010' /
                 'name_item.pickle').exists(),
                msg='existence of item in folder')

            self.assertEqual(
                json.dumps(xp._read_item('name_item', 10), sort_keys=True),
                json.dumps(self.item, sort_keys=True),
                msg='content of item')

            self.assertTrue(
                xp._read_item('name_item', 11) is None,
                msg='error if missing task folder')

            self.assertTrue(
                xp._read_item(
                    'unknown',
                    10, ) is None,
                msg='error if unknown item')

            xp._write_item('name_item', 4, self.item)
            xp._write_item('name_item', 7, self.item)
            xp._delete_item(11)
            xp._delete_item(10)
            self.assertTrue(
                not (Path(tmpdir) / 'tasks' / '000010').exists(),
                msg='supression of folder')
            xp._delete_item(7, 'wrong_name')
            self.assertTrue(
                (Path(tmpdir) / 'tasks' /
                 '000007'/'name_item.pickle').exists(),
                msg='supression of folder')
            xp._delete_item(7, 'name_item')
            self.assertTrue(
                (Path(tmpdir) / 'tasks' / '000007').exists(),
                msg='supression of folder')
            self.assertTrue(
                not (Path(tmpdir) / 'tasks' /
                     '000007'/'name_item.pickle').exists(),
                msg='supression of folder')

    def test_get_all_items(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp._write_item('name_item', 0, self.item)
            xp._write_item('name_item', 1, self.item)
            xp._write_item('name_item', 2, self.item)

            items = xp._get_all_items('name_item')

            self.assertTrue(
                isinstance(items, dict), msg='if output is dict')
            self.assertEqual(len(items), 3, msg='number of items')
            for idi, item in items.items():
                self.assertEqual(
                    item, self.item, msg='item {}'.format(idi))

            items = xp._get_all_items('name_item', only_ids=True)
            self.assertTrue(
                isinstance(items, list), msg='if output is list')
            self.assertEqual(len(items), 3, msg='number of items')
            self.assertEqual(set(items), set(range(3)), 'items')

    def test_add_tasks(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:

            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            # Add tasks
            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)

            schema = xp._schema
            self.assertTrue(isinstance(schema, dict), msg='type _schema')
            expected_schema = {('data', 'a'): [1, 2, 3],
                               ('data', 'b'): [4, 5],
                               ('data', 'seed'): [12],
                               ('problem', 'c'): [1, 2, 3],
                               ('problem', 'd'): [5],
                               ('problem', 'e'): ['alpha', 'beta'],
                               ('problem', 'seed'): [12],
                               ('solver', 'f'): [31],
                               ('solver', 'seed'): [12]}
            self.assertEqual(schema, expected_schema, msg='Value _schema')

            # Add new tasks
            new_data_params = {'a': [10, 11]}

            xp.add_tasks(new_data_params, self.problem_params,
                         self.solver_params)

            expected_schema = {('data', 'a'): [1, 2, 3, 10, 11],
                               ('data', 'b'): [4, 5],
                               ('data', 'seed'): [12],
                               ('problem', 'c'): [1, 2, 3],
                               ('problem', 'd'): [5],
                               ('problem', 'e'): ['alpha', 'beta'],
                               ('problem', 'seed'): [12],
                               ('solver', 'f'): [31],
                               ('solver', 'seed'): [12]}

            schema = xp._schema
            self.assertEqual(
                schema[('data', 'a')], [1, 2, 3, 10, 11], msg='update _schema')

            error_msg = 'data_params should be a dict'
            msg = 'error if wrong data params'
            with self.assertRaisesRegex(TypeError, error_msg, msg=msg):
                xp.add_tasks([2, 5], self.problem_params, self.solver_params)

            error_msg = 'problem_params should be a dict'
            msg = 'error if wrong problem params'
            with self.assertRaisesRegex(TypeError, error_msg, msg=msg):
                xp.add_tasks(self.data_params, [2, 5], self.solver_params)

            error_msg = 'solver_params should be a dict'
            msg = 'error if wrong solver params'
            with self.assertRaisesRegex(TypeError, error_msg, msg=msg):
                xp.add_tasks(self.data_params, self.problem_params, [2, 5])

            error_msg = 'wrong_param is not a parameter for getting data.'
            msg = 'error if wrong data param name'
            with self.assertRaisesRegex(ValueError, error_msg, msg=msg):
                xp.add_tasks({'wrong_param': 10}, self.problem_params,
                             self.solver_params)

            error_msg = 'wrong_param is not a parameter for getting the ' \
                        'problem.'
            msg = 'error if wrong problem param name'
            with self.assertRaisesRegex(ValueError, error_msg, msg=msg):
                xp.add_tasks(self.data_params, {'wrong_param': 10},
                             self.solver_params)

            error_msg = 'wrong_param is not a parameter for getting the ' \
                        'solver.'
            msg = 'error if wrong solver param name'
            with self.assertRaisesRegex(ValueError, error_msg, msg=msg):
                xp.add_tasks(self.data_params, self.problem_params,
                             {'wrong_param': 10})

    def test_generate_tasks(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:

            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)

            xp.generate_tasks()
            tasks = list(xp._get_all_items('task_params', only_ids=True))

            self.assertEqual(len(tasks), self.expected_nb_tasks,
                             msg='number of tasks')

            # Select a random task
            tasks = list(xp._get_all_items('task_params').values())
            idt = np.random.randint(len(tasks))
            task = tasks[idt]

            self.assertTrue(
                'data_params' in task, msg='data_params in task')
            self.assertTrue(
                task['data_params']['a'] in self.data_params['a'], msg='a')
            self.assertTrue(
                task['data_params']['b'] in self.data_params['b'], msg='b')

            self.assertTrue(
                'problem_params' in task, msg='problems_params in task')
            self.assertTrue(
                task['problem_params']['c'] in self.problem_params['c'],
                msg='c')
            self.assertTrue(
                task['problem_params']['d'] in self.problem_params['d'],
                msg='d')
            self.assertTrue(
                task['problem_params']['e'] in self.problem_params['e'],
                msg='e')

            self.assertTrue(
                'solver_params' in task, msg='solver_params in task')
            self.assertTrue(
                task['solver_params']['f'] in self.solver_params['f'],
                msg='e')

            xp.add_tasks({'a': 24}, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()
            tasks = list(xp._get_all_items('task_params', only_ids=True))

            all_values = list(self.data_params.values()) \
                + list(self.problem_params.values()) \
                + list(self.solver_params.values())
            all_len = [len(val) for val in all_values if isinstance(val, list)]
            all_len[0] += 1
            expected_nb_tasks = np.product(all_len)

            self.assertEqual(len(tasks), expected_nb_tasks,
                             msg='number of tasks')

    def test_run_task_by_id(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()
            xp.run_task_by_id(10)

            results = xp._get_all_items('result')
            self.assertTrue(10 in results, msg='the presence of a result')

            self.assertEqual(results[10], {'snr': 0.2, 'error': 1.6})

            # Rerun the task to check that already computed data are not
            # recomputed which is confirmed by checking the coverage report for
            # the corresponding part of the code.
            xp.run_task_by_id(10)

            error_msg = 'Unknown task: 100'
            msg = 'error if unknown task'
            with self.assertRaisesRegex(ValueError, error_msg, msg=msg):
                xp.run_task_by_id(100)

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)
            data_params = {'a': 1000, 'b': 0, 'seed': 12}
            problem_params = {'c': 1, 'd': 5, 'e': 'alpha', 'seed': 12}
            solver_params = {'f': 31, 'seed': 12}
            xp.add_tasks(data_params, problem_params, solver_params)
            xp.generate_tasks()
            error_msg = 'Exception in get_data for testing'
            with self.assertRaisesRegex(ValueError, error_msg):
                xp.run_task_by_id(0)

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)
            data_params = {'a': 1, 'b': 0, 'seed': 12}
            problem_params = {'c': 1000, 'd': 0, 'e': '', 'seed': 12}
            solver_params = {'f': 31, 'seed': 12}
            xp.add_tasks(data_params, problem_params, solver_params)
            xp.generate_tasks()
            error_msg = 'Exception in generate for testing'
            with self.assertRaisesRegex(ValueError, error_msg):
                xp.run_task_by_id(0)

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)
            data_params = {'a': 1, 'b': 0, 'seed': 12}
            problem_params = {'c': 1, 'd': 0, 'e': 'alpha', 'seed': 12}
            solver_params = {'f': 1000, 'seed': 12}
            xp.add_tasks(data_params, problem_params, solver_params)
            xp.generate_tasks()
            error_msg = 'Exception in solve for testing'
            with self.assertRaisesRegex(ValueError, error_msg):
                xp.run_task_by_id(0)

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)
            data_params = {'a': 10000, 'b': 0, 'seed': 12}
            problem_params = {'c': 1, 'd': 0, 'e': 'alpha', 'seed': 12}
            solver_params = {'f': 0, 'seed': 12}
            xp.add_tasks(data_params, problem_params, solver_params)
            xp.generate_tasks()
            error_msg = 'Exception in measure for testing'
            with self.assertRaisesRegex(ValueError, error_msg):
                xp.run_task_by_id(0)

        def get_data_failing(a, b, seed):
            raise ValueError('get_data failed')

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)
            get_data_original = xp.get_data
            xp.get_data = get_data_failing
            data_params = {'a': 1, 'b': 0, 'seed': 12}
            problem_params = {'c': 1, 'd': 5, 'e': 'alpha', 'seed': 12}
            solver_params = {'f': 31, 'seed': 12}
            xp.add_tasks(data_params, problem_params, solver_params)
            xp.generate_tasks()
            error_msg = 'get_data failed'
            with self.assertRaisesRegex(ValueError, error_msg):
                xp.run_task_by_id(0)
            self.assertTrue(xp._has_item('error_log', 0))
            xp.get_data = get_data_original
            xp.run_task_by_id(0)
            self.assertFalse(xp._has_item('error_log', 0))

        def get_data_wrong_output(a, b, seed):
            return {'Y': np.arange(a + b)}

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)
            xp.get_data = get_data_wrong_output
            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()
            error_msg = (r'Task 10, the output of get_data is not compatible '
                         r'with the input of the problem generated using '
                         r'get_problem. The parameter Y is unexpected.')
            with self.assertRaisesRegex(TypeError, error_msg):
                xp.run_task_by_id(10)

        class ProblemWrongOutput:
            def __init__(self, c, d, e, seed):
                pass

            def __call__(self, X):
                problem_data = {'Y': X}
                solution_data = dict()
                return problem_data, solution_data

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)
            xp.get_problem = ProblemWrongOutput
            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()
            error_msg = (r'Task 10, the output of the problem generated using '
                         r'get_problem is not compatible with the input of '
                         r'the solver generated using get_solver. '
                         r'The parameter Y is unexpected.')
            with self.assertRaisesRegex(TypeError, error_msg):
                xp.run_task_by_id(10)

    def test_launch_experiment(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()
            xp.launch_experiment()
            results = xp._get_all_items('result')

            self.assertEqual(len(results), self.expected_nb_tasks,
                             msg='the number of results')

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()
            xp.launch_experiment([2, 4, 8, 9])

            results = xp._get_all_items('result')
            self.assertEqual(len(results), 4, msg='the number of results')

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()
            xp.launch_experiment('1,3,7')

            results = xp._get_all_items('result')
            self.assertEqual(len(results), 3, msg='the number of results')

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()

            xp.launch_experiment([1, 100])

            results = xp._get_all_items('result')
            # 100 is not an id of the generated tasks, so only 1 result is
            # expected
            self.assertEqual(len(results), 1, msg='the number of results')

    def test_collect_results_full_xp(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:

            # launch experiment
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()

            values = list(xp._schema.values())

            error_msg = ('No result available. Results must be computed '
                         'before running collect_results().')
            with self.assertRaisesRegex(RuntimeError, error_msg):
                xp.collect_results()

            xp.launch_experiment()
            xp.collect_results()

            self.assertTrue(
                (xp.xp_path / 'results.npz').exists(),
                msg='the presence of file')

            file_npz = np.load(str(xp.xp_path / 'results.npz'))
            self.assertTrue('results' in file_npz, msg='the file')

            results = file_npz['results']
            self.assertTrue(
                isinstance(results, np.ndarray),
                msg='if results is nd-array')
            self.assertEqual(results.dtype, np.float64)
            self.assertEqual(
                len(results.shape), len(values) + 1, msg='shape of results')

            for ida, n_ticks in enumerate(results.shape[:-1]):
                self.assertEqual(
                    n_ticks,
                    len(values[ida]),
                    msg='size of dimension {}'.format(ida))

            measure_output = [0.2, 1.6]
            self.assertEqual(results.shape[-1], len(measure_output))

            self.assertTrue(np.all(results[..., 0] == measure_output[0]))
            self.assertTrue(np.all(results[..., 1] == measure_output[1]))

    def test_collect_results_partial_xp(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()

            values = list(xp._schema.values())

            xp.run_task_by_id(10)
            xp.run_task_by_id(20)
            xp.collect_results()

            self.assertTrue(
                (xp.xp_path / 'results.npz').exists(),
                msg='the presence of file')

            file_npz = np.load(str(xp.xp_path / 'results.npz'))
            self.assertTrue(
                'results' in file_npz, msg='if file contains results')

            results = file_npz['results']
            self.assertTrue(
                isinstance(results, np.ndarray),
                msg='if results is nd-array')
            self.assertEqual(results.dtype, np.float64)
            self.assertEqual(
                len(results.shape), len(values) + 1, msg='shape of results')

            for ida, n_ticks in enumerate(results.shape[:-1]):
                self.assertEqual(
                    n_ticks,
                    len(values[ida]),
                    msg='size of dimension {}'.format(ida))

            measure_output = [0.2, 1.6]
            self.assertEqual(results.shape[-1], len(measure_output))

            subarray = results[..., 0]
            self.assertTrue(
                np.all(subarray[~np.isnan(subarray)] == measure_output[0]))
            subarray = results[..., 1]
            self.assertTrue(
                np.all(subarray[~np.isnan(subarray)] == measure_output[1]))

    def test_load_results(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()

            values = list(xp._schema.values())

            xp.run_task_by_id(10)
            xp.run_task_by_id(20)

            error_msg = ('No file with collected results available. '
                         'Results must be computed and collected before '
                         'running load_results().')
            with self.assertRaisesRegex(RuntimeError, error_msg):
                xp.load_results()

            xp.collect_results()

            results, axes_labels, axes_values = xp.load_results()

            self.assertTrue(
                isinstance(results, np.ndarray),
                msg='if results is nd-array')
            self.assertEqual(results.dtype, np.float64)
            self.assertEqual(
                len(results.shape), len(values) + 1, msg='shape of results')

            for ida, n_ticks in enumerate(results.shape[:-1]):
                self.assertEqual(
                    n_ticks,
                    len(values[ida]),
                    msg='size of dimension {}'.format(ida))
                np.testing.assert_array_equal(values[ida], axes_values[ida])

            measure_output = [0.2, 1.6]
            self.assertEqual(results.shape[-1], len(measure_output))

            subarray = results[..., 0]
            self.assertTrue(
                np.all(subarray[~np.isnan(subarray)] == measure_output[0]))
            subarray = results[..., 1]
            self.assertTrue(
                np.all(subarray[~np.isnan(subarray)] == measure_output[1]))

            expected_labels = ('data_a', 'data_b', 'data_seed', 'problem_c',
                               'problem_d', 'problem_e', 'problem_seed',
                               'solver_f', 'solver_seed', 'measure')
            np.testing.assert_array_equal(axes_labels, expected_labels)

            error_msg = ('Unrecognized value for array_type, the valid '
                         'options are "numpy" and "xarray".')
            with self.assertRaisesRegex(ValueError, error_msg):
                _ = xp.load_results('wrong_flag')

            error_msg = (r'Please install the xarray package to '
                         r'run the load_results\(\) method with the '
                         r'array\_type="xarray" option.')
            with self.assertRaisesRegex(ImportError, error_msg):
                with patch.dict(sys.modules, {'xarray': None}):
                    _ = xp.load_results('xarray')

    def test_load_results_as_xarray(self):
        try:
            import xarray
        except ImportError:
            msg = ("As the xarray package is not available, the "
                   "array_type='xarray' option of Experiment.load_results() "
                   "could not be tested.")
            warnings.warn(msg)
            return

        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()

            values = list(xp._schema.values())

            xp.run_task_by_id(10)
            xp.run_task_by_id(20)

            xp.collect_results()

            results = xp.load_results('xarray')

            self.assertTrue(
                isinstance(results, xarray.DataArray),
                msg='if results is nd-array')
            self.assertEqual(results.dtype, np.float64)
            self.assertEqual(
                len(results.shape), len(values) + 1, msg='shape of results')

            for ida, n_ticks in enumerate(results.shape[:-1]):
                self.assertEqual(
                    n_ticks,
                    len(values[ida]),
                    msg='size of dimension {}'.format(ida))
                np.testing.assert_array_equal(
                    values[ida], results.coords[results.dims[ida]])

            measure_output = [0.2, 1.6]
            self.assertEqual(results.shape[-1], len(measure_output))

            subarray = results[..., 0].values
            self.assertTrue(
                np.all(subarray[~np.isnan(subarray)] == measure_output[0]))
            subarray = results[..., 1].values
            self.assertTrue(
                np.all(subarray[~np.isnan(subarray)] == measure_output[1]))

            expected_labels = ('data_a', 'data_b', 'data_seed', 'problem_c',
                               'problem_d', 'problem_e', 'problem_seed',
                               'solver_f', 'solver_seed', 'measure')
            self.assertEqual(results.dims, expected_labels)

    def test_display_status(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            data_params = {'a': [1000, 2, 3], 'b': [4, 5], 'seed': 12}

            xp.add_tasks(data_params, self.problem_params, self.solver_params)
            xp.generate_tasks()

            nb_tasks = self.expected_nb_tasks

            nb_tasks_success = 7
            xp.launch_experiment(range(nb_tasks_success))

            nb_tasks_fail = 3
            xp.launch_experiment(range(nb_tasks - nb_tasks_fail, nb_tasks))

            tmp_stdout = StringIO()
            with contextlib.redirect_stdout(tmp_stdout):
                xp.display_status()
            out = tmp_stdout.getvalue().strip()
            expected_out = 'Number of tasks: {}\n' \
                           'Number of finished tasks: {}\n' \
                           'Number of failed tasks: {}\n' \
                           'Number of pending tasks: {}'
            expected_out = expected_out.format(
                nb_tasks, nb_tasks_success, nb_tasks_fail,
                nb_tasks - nb_tasks_success)
            self.assertEqual(out, expected_out)

            tmp_stdout = StringIO()
            with contextlib.redirect_stdout(tmp_stdout):
                xp.display_status(False)
            out = tmp_stdout.getvalue().strip()
            expected_out = ''
            for task_id in range(nb_tasks_success):
                expected_out += '{:06d}..................OK\n'.format(task_id)
            for task_id in range(nb_tasks_success, nb_tasks - nb_tasks_fail):
                expected_out += '{:06d}......NOT CALCULATED\n'.format(task_id)
            for task_id in range(nb_tasks - nb_tasks_fail, nb_tasks):
                expected_out += '{:06d}...............ERROR\n'.format(task_id)
            expected_out = expected_out[:-1]
            self.assertEqual(out, expected_out)

    def test_get_task_data_by_id(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()
            task_data = xp.get_task_data_by_id(0)

            expected_dict_keys = {'id_task', 'task_params', 'source_data',
                                  'problem_data', 'solution_data',
                                  'solved_data', 'result'}

            expected = {'id_task': 0,
                        'task_params': {
                            'data_params': {'a': 1, 'b': 4, 'seed': 12},
                            'problem_params':
                                {'c': 1, 'd': 5, 'e': 'alpha', 'seed': 12},
                            'solver_params': {'f': 31, 'seed': 12}},
                        'source_data': None,
                        'problem_data': None,
                        'solved_data': None,
                        'solution_data': None,
                        'result': None}

            self.assertEqual(expected_dict_keys, set(task_data.keys()))
            self.assertEqual(expected, task_data)

            xp.run_task_by_id(0)
            task_data = xp.get_task_data_by_id(0)

            expected = {'id_task': 0,
                        'task_params': {
                            'data_params': {'a': 1, 'b': 4, 'seed': 12},
                            'problem_params':
                                {'c': 1, 'd': 5, 'e': 'alpha', 'seed': 12},
                            'solver_params': {'f': 31, 'seed': 12}},
                        'source_data': {'X': np.array([0, 1, 2, 3, 4])},
                        'problem_data': {'X': np.array([11, 12, 13, 14, 15])},
                        'solution_data': {'X': np.array([12, 13, 14, 15, 16])},
                        'solved_data': {'X': np.array([42, 43, 44, 45, 46])},
                        'result': {'snr': 0.2, 'error': 1.6}}

            self.assertEqual(expected_dict_keys, set(task_data.keys()))
            self.assertEqual(expected['id_task'], task_data['id_task'])
            self.assertEqual(expected['task_params'], task_data['task_params'])
            np.testing.assert_equal(expected['source_data']['X'],
                                    task_data['source_data']['X'])
            np.testing.assert_equal(expected['problem_data']['X'],
                                    task_data['problem_data']['X'])
            np.testing.assert_equal(expected['solution_data']['X'],
                                    task_data['solution_data']['X'])
            np.testing.assert_equal(expected['solved_data']['X'],
                                    task_data['solved_data']['X'])
            self.assertEqual(expected['result'], task_data['result'])

            error_msg = 'Unknown task: 100'
            with self.assertRaisesRegex(ValueError, error_msg):
                _ = xp.get_task_data_by_id(100)

    def test_get_task_data_by_params(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()
            task_params = {'data_params': {'a': 1, 'b': 4, 'seed': 12},
                           'problem_params':
                               {'c': 1, 'd': 5, 'e': 'alpha', 'seed': 12},
                           'solver_params': {'f': 31, 'seed': 12}}
            task_data = xp.get_task_data_by_params(**task_params)

            expected_dict_keys = {'id_task', 'task_params', 'source_data',
                                  'problem_data', 'solution_data',
                                  'solved_data', 'result'}

            expected = {'id_task': 0,
                        'task_params': task_params,
                        'source_data': None,
                        'problem_data': None,
                        'solution_data': None,
                        'solved_data': None,
                        'result': None}

            self.assertEqual(expected_dict_keys, set(task_data.keys()))
            self.assertEqual(expected, task_data)

            xp.run_task_by_id(0)
            task_data = xp.get_task_data_by_params(**task_params)

            expected = {'id_task': 0,
                        'task_params': task_params,
                        'source_data': {'X': np.array([0, 1, 2, 3, 4])},
                        'problem_data': {'X': np.array([11, 12, 13, 14, 15])},
                        'solution_data': {'X': np.array([12, 13, 14, 15, 16])},
                        'solved_data': {'X': np.array([42, 43, 44, 45, 46])},
                        'result': {'snr': 0.2, 'error': 1.6}}

            self.assertEqual(expected_dict_keys, set(task_data.keys()))
            self.assertEqual(expected['id_task'], task_data['id_task'])
            self.assertEqual(expected['task_params'], task_data['task_params'])
            np.testing.assert_equal(expected['source_data']['X'],
                                    task_data['source_data']['X'])
            np.testing.assert_equal(expected['problem_data']['X'],
                                    task_data['problem_data']['X'])
            np.testing.assert_equal(expected['solution_data']['X'],
                                    task_data['solution_data']['X'])
            np.testing.assert_equal(expected['solved_data']['X'],
                                    task_data['solved_data']['X'])
            self.assertEqual(expected['result'], task_data['result'])

            task_params = {'data_params': {'a': 100, 'b': 4, 'seed': 12},
                           'problem_params':
                               {'c': 1, 'd': 5, 'e': 'alpha', 'seed': 12},
                           'solver_params': {'f': 31, 'seed': 12}}
            error_msg = 'No task exists for the given parameters.'
            with self.assertRaisesRegex(ValueError, error_msg):
                _ = xp.get_task_data_by_params(**task_params)

            error_msg = 'Parameters must be of type dict.'
            with self.assertRaisesRegex(TypeError, error_msg):
                _ = xp.get_task_data_by_params(12, {'c': 1}, {'f': 31})

            task_params = {'data_params': {'b': 4, 'seed': 12},
                           'problem_params':
                               {'c': 1, 'd': 5, 'e': 'alpha', 'seed': 12},
                           'solver_params': {'f': 31, 'seed': 12}}
            error_msg = ('Wrong keys in data_params, the expected keys are: '
                         'a, b, seed.')
            with self.assertRaisesRegex(ValueError, error_msg):
                _ = xp.get_task_data_by_params(**task_params)

            task_params = {'data_params': {'a': 100, 'b': 4, 'seed': 12},
                           'problem_params':
                               {'c': 1, 'd': 5, 'f': 'alpha', 'seed': 12},
                           'solver_params': {'f': 31, 'seed': 12}}
            error_msg = ('Wrong keys in problem_params, the expected keys are:'
                         ' c, d, e, seed.')
            with self.assertRaisesRegex(ValueError, error_msg):
                _ = xp.get_task_data_by_params(**task_params)

            task_params = {'data_params': {'a': 100, 'b': 4, 'seed': 12},
                           'problem_params':
                               {'c': 1, 'd': 5, 'e': 'alpha', 'seed': 12},
                           'solver_params': {'e': 31}}
            error_msg = ('Wrong keys in solver_params, the expected keys are:'
                         ' f, seed.')
            with self.assertRaisesRegex(ValueError, error_msg):
                _ = xp.get_task_data_by_params(**task_params)

    def test_track_tasks(self):
        with tempfile.TemporaryDirectory(dir=str(get_data_path())) as tmpdir:

            xp_name = Path(tmpdir).stem
            Path(tmpdir).rmdir()

            exp = Experiment(name=xp_name,
                             get_data=MockTrackData(),
                             get_problem=MockTrackProblem,
                             get_solver=MockTrackSolver,
                             measure=mock_track_measure)
            exp.reset()
            exp.add_tasks(data_params=MockTrackData.get_params(),
                          problem_params=MockTrackProblem.get_params(),
                          solver_params=MockTrackSolver.get_params())
            exp.generate_tasks()
            exp.launch_experiment()
            exp.collect_results()
            results, axes_labels, axes_values = \
                exp.load_results(array_type='numpy')

            check_tracked_results(results=results,
                                  axes_labels=axes_labels,
                                  axes_values=axes_values)

            exp.add_tasks(
                data_params=MockTrackData.get_additional_params(),
                problem_params=MockTrackProblem.get_additional_params(),
                solver_params=MockTrackSolver.get_additional_params())
            exp.generate_tasks()
            exp.launch_experiment()
            exp.collect_results()
            results, axes_labels, axes_values = \
                exp.load_results(array_type='numpy')
            check_tracked_results(results=results,
                                  axes_labels=axes_labels,
                                  axes_values=axes_values)

    def test_experiment_with_specified_data_path(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            xp_name = Path(tmpdir).stem
            xp = MockExperiment(xp_name, data_path=tmpdir)

            xp.add_tasks(self.data_params, self.problem_params,
                         self.solver_params)
            xp.generate_tasks()
            xp.run_task_by_id(10)

            results = xp._get_all_items('result')
            self.assertTrue(10 in results, msg='the presence of a result')


# Test using a more realistic example based on the yafe tutorial

def get_sine_data(f0, signal_len=1000):
    return {'signal': np.sin(2*np.pi*f0*np.arange(signal_len))}


class SimpleDenoisingProblem:
    def __init__(self, snr_db):
        self.snr_db = snr_db

    def __call__(self, signal):
        random_state = np.random.RandomState(0)
        noise = random_state.randn(*signal.shape)
        observation = signal + 10 ** (-self.snr_db / 20) * noise \
            / np.linalg.norm(noise) * np.linalg.norm(signal)
        problem_data = {'observation': observation}
        solution_data = {'signal': signal}
        return problem_data, solution_data

    def __str__(self):
        return 'SimpleDenoisingProblem(snr_db={})'.format(self.snr_db)


class SmoothingSolver:
    def __init__(self, filter_len):
        self.filter_len = filter_len

    def __call__(self, observation):
        smoothing_filter = np.hamming(self.filter_len)
        smoothing_filter /= np.sum(smoothing_filter)
        return {'reconstruction': np.convolve(observation, smoothing_filter,
                                              mode='same')}

    def __str__(self):
        return 'SmoothingSolver(filter_len={})'.format(self.filter_len)


def measure_full_example(solution_data, solved_data, task_params=None,
                         source_data=None, problem_data=None):
    euclidian_distance = np.linalg.norm(solution_data['signal']
                                        - solved_data['reconstruction'])
    sdr = 20 * np.log10(np.linalg.norm(solution_data['signal'])
                        / euclidian_distance)
    inf_distance = np.linalg.norm(solution_data['signal']
                                  - solved_data['reconstruction'], ord=np.inf)
    return {'sdr': sdr,
            'euclidian_distance': euclidian_distance,
            'inf_distance': inf_distance}


class TestFullExample(unittest.TestCase):

    def setUp(self):
        self.temp_data_path = tempfile.mkdtemp(prefix='yafe_')

    def tearDown(self):
        rmtree(self.temp_data_path)

    def test_full_example(self):
        decimal_precision = 13
        data_params = {'f0': np.arange(0.01, 0.1, 0.01), 'signal_len': [1000]}
        problem_params = {'snr_db': [-10, 0, 30]}
        solver_params = {'filter_len': 2**np.arange(6, step=2)}
        exp = Experiment(name='full_experiment',
                         get_data=get_sine_data,
                         get_problem=SimpleDenoisingProblem,
                         get_solver=SmoothingSolver,
                         measure=measure_full_example,
                         data_path=self.temp_data_path,
                         log_to_file=False,
                         log_to_console=False)
        exp.add_tasks(data_params=data_params,
                      problem_params=problem_params,
                      solver_params=solver_params)
        status = ('Number of tasks: 0\n'
                  'Number of finished tasks: 0\n'
                  'Number of failed tasks: 0\n'
                  'Number of pending tasks: 0\n')
        self.assertEqual(status, exp._get_status(True))
        exp.generate_tasks()
        status = ('Number of tasks: 81\n'
                  'Number of finished tasks: 0\n'
                  'Number of failed tasks: 0\n'
                  'Number of pending tasks: 81\n')
        self.assertEqual(status, exp._get_status(True))
        exp.run_task_by_id(idt=5)
        status = ('Number of tasks: 81\n'
                  'Number of finished tasks: 1\n'
                  'Number of failed tasks: 0\n'
                  'Number of pending tasks: 80\n')
        self.assertEqual(status, exp._get_status(True))
        exp.launch_experiment()
        status = ('Number of tasks: 81\n'
                  'Number of finished tasks: 81\n'
                  'Number of failed tasks: 0\n'
                  'Number of pending tasks: 0\n')
        self.assertEqual(status, exp._get_status(True))
        exp.collect_results()
        results, axes_labels, axes_values = exp.load_results()
        self.assertEqual((9, 1, 3, 3, 3), results.shape)
        self.assertEqual(np.float64, results.dtype)
        self.assertAlmostEqual(10.100157318786618, np.mean(results),
                               places=decimal_precision)
        expected_labels = np.array(
            ['data_f0', 'data_signal_len', 'problem_snr_db',
             'solver_filter_len', 'measure'],
            dtype=object)
        np.testing.assert_array_equal(expected_labels, axes_labels)
        expected_values = np.array(
            [
             np.array([0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09]),
             np.array([1000]),
             np.array([-10, 0, 30]),
             np.array([1,  4, 16]),
             np.array(['sdr', 'euclidian_distance', 'inf_distance'],
                      dtype=object)
            ],
            dtype=object)
        np.testing.assert_array_almost_equal(
            expected_values[0], axes_values[0], decimal=decimal_precision)
        for ind in range(1, 5):
            np.testing.assert_array_equal(expected_values[ind],
                                          axes_values[ind])
        additional_snr_db = [10, 20]
        additional_filter_len = 2**np.arange(6)
        exp.add_tasks(problem_params={'snr_db': additional_snr_db},
                      data_params=dict(),
                      solver_params={'filter_len': additional_filter_len})
        exp.generate_tasks()
        status = ('Number of tasks: 270\n'
                  'Number of finished tasks: 81\n'
                  'Number of failed tasks: 0\n'
                  'Number of pending tasks: 189\n')
        self.assertEqual(status, exp._get_status(True))
        exp.launch_experiment()
        status = ('Number of tasks: 270\n'
                  'Number of finished tasks: 270\n'
                  'Number of failed tasks: 0\n'
                  'Number of pending tasks: 0\n')
        self.assertEqual(status, exp._get_status(True))
        exp.collect_results()
        results, axes_labels, axes_values = exp.load_results()
        self.assertEqual((9, 1, 5, 6, 3), results.shape)
        self.assertEqual(np.float64, results.dtype)
        self.assertAlmostEqual(8.2145574032230524, np.mean(results),
                               places=decimal_precision)
        np.testing.assert_array_equal(expected_labels, axes_labels)
        expected_values = np.array(
            [
             np.array([0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09]),
             np.array([1000]),
             np.array([-10, 0, 10, 20, 30]),
             np.array([1, 2, 4, 8, 16, 32]),
             np.array(['sdr', 'euclidian_distance', 'inf_distance'],
                      dtype=object)
            ],
            dtype=object)
        np.testing.assert_array_almost_equal(
            expected_values[0], axes_values[0], decimal=decimal_precision)
        for ind in range(1, 5):
            np.testing.assert_array_equal(expected_values[ind],
                                          axes_values[ind])



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# yafe: Yet Another Framework for Experiments.
#
# Licence
# -------
# This file is part of yafe.
#
# yafe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Example script to test yafe.utils.generate_oar_script()

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
.. moduleauthor:: Florent Jaillet
"""
from pathlib import Path
import tempfile

import numpy as np

from yafe import Experiment


def get_sine_data(f0, signal_len=1000):
    return {'signal': np.sin(2*np.pi*f0*np.arange(signal_len))}


class SimpleDenoisingProblem:
    def __init__(self, snr_db):
        self.snr_db = snr_db

    def __call__(self, signal):
        random_state = np.random.RandomState(0)
        noise = random_state.randn(*signal.shape)
        observation = signal + 10 ** (-self.snr_db / 20) * noise \
            / np.linalg.norm(noise) * np.linalg.norm(signal)
        problem_data = {'observation': observation}
        solution_data = {'signal': signal}
        return problem_data, solution_data

    def __str__(self):
        return 'SimpleDenoisingProblem(snr_db={})'.format(self.snr_db)


class SmoothingSolver:
    def __init__(self, filter_len):
        self.filter_len = filter_len

    def __call__(self, observation):
        smoothing_filter = np.hamming(self.filter_len)
        smoothing_filter /= np.sum(smoothing_filter)
        return {'reconstruction': np.convolve(observation, smoothing_filter,
                                              mode='same')}

    def __str__(self):
        return 'SmoothingSolver(filter_len={})'.format(self.filter_len)


def measure(solution_data, solved_data, task_params=None, source_data=None,
            problem_data=None):
    euclidian_distance = np.linalg.norm(solution_data['signal']
                                        - solved_data['reconstruction'])
    sdr = 20 * np.log10(np.linalg.norm(solution_data['signal'])
                        / euclidian_distance)
    inf_distance = np.linalg.norm(solution_data['signal']
                                  - solved_data['reconstruction'], ord=np.inf)
    return {'sdr': sdr,
            'euclidian_distance': euclidian_distance,
            'inf_distance': inf_distance}


def create_tasks(exp):
    data_params = {'f0': np.arange(0.01, 0.1, 0.01), 'signal_len': [1000]}
    problem_params = {'snr_db': [-10, 0, 30]}
    solver_params = {'filter_len': 2**np.arange(6, step=2)}
    exp.add_tasks(data_params=data_params,
                  problem_params=problem_params,
                  solver_params=solver_params)
    exp.generate_tasks()


temp_data_path = tempfile.gettempdir() / Path('yafe_temp_test_data')
if not temp_data_path.exists():
    temp_data_path.mkdir()


experiment = Experiment(name='test_generate_oar_script',
                        get_data=get_sine_data,
                        get_problem=SimpleDenoisingProblem,
                        get_solver=SmoothingSolver,
                        measure=measure,
                        data_path=temp_data_path,
                        log_to_file=False,
                        log_to_console=False)


if __name__ == '__main__':
    create_tasks(experiment)



# -*- coding: utf-8 -*-
# ######### COPYRIGHT #########
#
# Copyright(c) 2018
# -----------------
#
# * Laboratoire d'Informatique et Systèmes <http://www.lis-lab.fr/>
# * Université d'Aix-Marseille <http://www.univ-amu.fr/>
# * Centre National de la Recherche Scientifique <http://www.cnrs.fr/>
# * Université de Toulon <http://www.univ-tln.fr/>
#
# Contributors
# ------------
#
# * Ronan Hamon <firstname.lastname_AT_lis-lab.fr>
# * Valentin Emiya <firstname.lastname_AT_lis-lab.fr>
# * Florent Jaillet <firstname.lastname_AT_lis-lab.fr>
#
# Description
# -----------
#
# yafe: Yet Another Framework for Experiments.
#
# Licence
# -------
# This file is part of yafe.
#
# yafe is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ######### COPYRIGHT #########
"""Test of the module :module:`yafe.utils`

.. moduleauthor:: Ronan Hamon
.. moduleauthor:: Valentin Emiya
.. moduleauthor:: Florent Jaillet
"""
from contextlib import redirect_stdout
from io import StringIO
import json
import logging
import os
from pathlib import Path
from shutil import rmtree
import runpy
import stat
import tempfile
import unittest

import numpy as np

from yafe.utils import get_logger, ConfigParser, generate_oar_script
from .test_base import create_temp_test_config, get_data_path, MockExperiment


class TestGetLogger(unittest.TestCase):

    def setUp(self):
        self._temp_dir = create_temp_test_config()

    def tearDown(self):
        rmtree(self._temp_dir)

    def test_get_logger(self):

        logger_path = ConfigParser().get_path('LOGGER', 'PATH', False)

        # no saving to file
        logger = get_logger(name='test', to_file=False, to_console=False)

        logger.info('test')
        self.assertEqual(logging.getLevelName(logger.level), 'DEBUG')
        self.assertEqual(len(logger.handlers), 0)

        # saving to file with path from config file
        self.assertFalse(logger_path.exists())

        logger = get_logger(name='test')

        self.assertTrue(logger_path.exists())

        self.assertEqual(logging.getLevelName(logger.level), 'DEBUG')
        self.assertEqual(len(logger.handlers), 2)

        self.assertEqual(
            logging.getLevelName(logger.handlers[0].level), 'DEBUG')
        self.assertEqual(logger.handlers[0].stream.name, str(logger_path))

        self.assertEqual(
            logging.getLevelName(logger.handlers[1].level), 'INFO')

        # saving to file with specified path
        tmp_file = tempfile.NamedTemporaryFile(delete=False)
        tmp_file.close()
        try:
            logger = get_logger(name='test',
                                logger_path=tmp_file.name)

            self.assertEqual(logging.getLevelName(logger.level), 'DEBUG')
            self.assertEqual(len(logger.handlers), 2)

            self.assertEqual(
                logging.getLevelName(logger.handlers[0].level), 'DEBUG')
            self.assertEqual(logger.handlers[0].stream.name, tmp_file.name)

            self.assertEqual(
                logging.getLevelName(logger.handlers[1].level), 'INFO')
        finally:
            os.remove(tmp_file.name)


class TestConfigParser(unittest.TestCase):

    def setUp(self):
        self._temp_dir = create_temp_test_config()

    def tearDown(self):
        rmtree(self._temp_dir)

    def test_generate_config_and_get_path(self):
        original_config_path = ConfigParser._config_path
        temp_dir = tempfile.mkdtemp(prefix='yafe_')
        try:
            ConfigParser._config_path = temp_dir / Path('yafe.log')
            ConfigParser().generate_config()

            # Check that the config file has been created
            self.assertTrue(ConfigParser._config_path.exists())

            config = ConfigParser()
            # Check that the USER section exists with an empty data_path
            user_data_path = config.get_path('USER', 'data_path')
            self.assertIs(user_data_path, None)
            # Check that the LOGGER section exists with an empty path
            logger_path = config.get_path('LOGGER', 'path')
            self.assertIs(logger_path, None)

            # Further check that get_path() works as expected
            config['TEST_SECTION'] = {'good_path': temp_dir,
                                      'bad_path': '/non_existing_path'}
            with ConfigParser._config_path.open('w') as configfile:
                config.write(configfile)
            good_path = config.get_path('TEST_SECTION', 'good_path')
            self.assertEqual(temp_dir, str(good_path))
            with self.assertRaises(IOError):
                bad_path = config.get_path('TEST_SECTION', 'bad_path')

        finally:
            ConfigParser._config_path = original_config_path
            rmtree(temp_dir)


class TestGenerateOarScript(unittest.TestCase):

    def tearDown(self):
        rmtree(tempfile.gettempdir() / Path('yafe_temp_test_data'))

    def test_generate_oar_script(self):
        script_file_path = (Path(__file__).parent
                            / 'script_for_generate_oar_script.py')
        xp_var_name = 'experiment'
        temp_dir = tempfile.gettempdir()
        xp_path = (temp_dir / Path('yafe_temp_test_data')
                   / 'test_generate_oar_script')

        runpy.run_path(str(script_file_path), run_name='__main__')

        tmp_stdout = StringIO()
        with redirect_stdout(tmp_stdout):
            generate_oar_script(script_file_path, xp_var_name)

        out = tmp_stdout.getvalue().strip()
        expected_out = ('oarsub -S {}/yafe_temp_test_data/test_generate_oar'
                        '_script/script_oar.sh'.format(temp_dir))
        self.assertEqual(out, expected_out)

        task_file_path = xp_path / "listoftasks.txt"
        self.assertTrue(task_file_path.exists())
        task_data = np.loadtxt(task_file_path, dtype=np.int64)
        np.testing.assert_array_equal(np.arange(81), task_data)

        tasks = [1, 3]
        generate_oar_script(script_file_path, xp_var_name, tasks)

        task_file_path = xp_path / "listoftasks.txt"
        self.assertTrue(task_file_path.exists())
        task_data = np.loadtxt(task_file_path, dtype=np.int64)
        np.testing.assert_array_equal(tasks, task_data)

        script_path = xp_path / "script_oar.sh"
        self.assertTrue(script_path.exists())
        # Check that the script is executable
        self.assertTrue(os.stat(script_path).st_mode & stat.S_IXUSR)
        # Check the content of the script
        with script_path.open() as script_file:
            script_content = script_file.readlines()
        self.assertEqual(len(script_content), 13)
        expected_content = [
            '#!/bin/sh\n',
            '#OAR -n test_generate_oar_script\n',
            '#OAR --array-param-file {}/yafe_temp_test_data/'
            'test_generate_oar_script/listoftasks.txt\n'.format(temp_dir),
            '#OAR -O {}/yafe_temp_test_data/'
            'test_generate_oar_script/%jobid%.out\n'.format(temp_dir),
            '#OAR -E {}/yafe_temp_test_data/'
            'test_generate_oar_script/%jobid%.err\n'.format(temp_dir),
            '#OAR -l walltime=02:00:00\n',
            '#OAR -p gpu IS NULL\n',
            'echo "OAR_JOB_ID: $OAR_JOB_ID"\n',
            'echo "OAR_ARRAY_ID: $OAR_ARRAY_ID"\n',
            'echo "OAR_ARRAY_INDEX: $OAR_ARRAY_INDEX"\n',
            'echo "Running experiment.launch_experiment(task_ids=\'$1\')"\n',
            'python -c "import sys; sys.path.append(\'{}\'); '
            'from script_for_generate_oar_script import experiment; '
            'experiment.launch_experiment(task_ids=\'$1\')"\n'.format(
                Path(__file__).parent),
            'exit $?']
        self.assertEqual(script_content, expected_content)

        activate_env_command = 'source activate test_env'
        generate_oar_script(script_file_path, xp_var_name,
                            activate_env_command=activate_env_command)
        self.assertTrue(script_path.exists())
        with script_path.open() as script_file:
            script_content = script_file.readlines()
        self.assertEqual(len(script_content), 14)
        expected_content.insert(10, '{}\n'.format(activate_env_command))
        self.assertEqual(script_content, expected_content)

        generate_oar_script(script_file_path, xp_var_name,
                            use_gpu=True)
        self.assertTrue(script_path.exists())
        with script_path.open() as script_file:
            script_content = script_file.readlines()
        self.assertEqual(len(script_content), 13)
        expected_content.pop(10)
        expected_content[6] = '#OAR -p gpu IS NOT NULL\n'
        self.assertEqual(script_content, expected_content)
